module expansion
   use types; use consts_dp
   use rad_tools; use emsn_tools
   use masters
   use pdfs_tools; use pdf_expansions
   use hoppet_v1
   use interpolation_radish
   use reader

   implicit none

   private
   public :: init_expansion, reset_dynamical_expansion, terminate_expansion
   public :: expanded_sigma, expanded_sigma_inclusive
   public :: expanded_sigma_ptj, expanded_sigma_ptj_inclusive

   type(PDFExpansion), public     :: pdf1e, pdf2e

contains

   !-----------------------------------------------------------------------
   subroutine init_expansion()
      call init_expansion_pdf()
      ! initialise numerical arrays containing the master integrals for the expansion
      call init_master_integrals()
   end subroutine init_expansion
   !-----------------------------------------------------------------------

   !-----------------------------------------------------------------------
   subroutine init_expansion_pdf()
      real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
      real(dp) :: u
      ! prepare the material needed for the PDFs
      call get_pdfs(cs%muF, cs%collider, pdf1, pdf2)
      !------------------------------------------------------------
      !------------------------------------------------------------
      call AllocPDFExpansion(pdf1e)
      call AllocPDFExpansion(pdf2e)
      !-- calculate expansion ---------------------------

      u = cs%ln_muF2_M2 - cs%ln_muR2_M2

      pdf1e%p = pdf1
      call set_pdf_expansion(pdf1e,u)
      pdf2e%p = pdf2
      call set_pdf_expansion(pdf2e,u)

   end subroutine init_expansion_pdf
   !-----------------------------------------------------------------------

   !-----------------------------------------------------------------------
   subroutine reset_dynamical_expansion()
      real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
      real(dp) :: u
      ! prepare the material needed for the PDFs
      call get_pdfs(cs%muF, cs%collider, pdf1, pdf2)
      !------------------------------------------------------------
      !-- calculate expansion ---------------------------

      u = cs%ln_muF2_M2 - cs%ln_muR2_M2

      pdf1e%p = pdf1
      call set_pdf_expansion(pdf1e,u)
      pdf2e%p = pdf2
      call set_pdf_expansion(pdf2e,u)

   end subroutine reset_dynamical_expansion
   !-----------------------------------------------------------------------

   !-----------------------------------------------------------------------
   subroutine expanded_sigma(histo_index, v, order, msqB, msqV1, msqV2, x1, x2, resummed_expanded, lumi_asympt)
      real(dp),                  intent(in) :: v(:)
      real(dp), intent(in) :: x1, x2
      integer,                   intent(in) :: histo_index, order !(0=LL,1=NLL,2=NNLL)
      real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int),msqV1(-nf_int:nf_int,-nf_int:nf_int),msqV2(-nf_int:nf_int,-nf_int:nf_int)
      ! first index of res denotes the order in alphas
      real(dp), intent(out) :: resummed_expanded(0:3,size(v))
      real(dp), optional, intent(out) :: lumi_asympt(0:3)
      !-------------------------------------
      type(PDFExpansion_at_x)  :: pdf1eatx, pdf2eatx
      type(LumiExpansion_at_x) :: lumi, luminll, luminnll

      if (order < order_NLL) return ! use standard resummation for LL

      ! evaluate the pdf expansions at the Bjorken variables given by the Born event
      call set_pdf_expansion_at_x(pdf1eatx, pdf2eatx, pdf1e, pdf2e, x1, x2)

      !--- the following call returns luminosities factors lij, where lij is
      !--- the coefficient of (as2pi(muR))^i * t^j with  t = 2*log(muF1/muF0)=-2*L
      call set_lumi_expansion_Born(pdf1eatx, pdf2eatx, msqB, msqV1, msqV2, lumi, order)

      if (order > order_NLL) then
         call set_lumi_expansion_Born(pdf1eatx, pdf2eatx, msqB, msqV1, msqV2, luminll, 1)
      end if

      if (order > order_NNLL) then
         call set_lumi_expansion_Born(pdf1eatx, pdf2eatx, msqB, msqV1, msqV2, luminll, 1)
         call set_lumi_expansion_Born(pdf1eatx, pdf2eatx, msqB, msqV1, msqV2, luminnll, 2)
      end if

      if (present(lumi_asympt)) then
         lumi_asympt = zero ! asymptotic limit for the multiplicative matching
         ! For multiplicative matching scheme, return asymptotic normalization
         if (order == order_NLL)  lumi_asympt(0)  = lumi_NLL_Born(cs%muF, zero, x1, x2, msqB, msqV1)
         if (order == order_NNLL) lumi_asympt(0)  = lumi_NNLL_Born(cs%muF, zero, x1, x2, msqB, msqV1, msqV2)
         if (order == order_NNNLL) lumi_asympt(0) = lumi_NNNLL_Born(cs%muF, zero, x1, x2, msqB, msqV1, msqV2)
         lumi_asympt(1) = lumi%l10 * (cs%alphas_muR/twopi)
         lumi_asympt(2) = lumi%l20 * (cs%alphas_muR/twopi)**2
         lumi_asympt(3) = lumi%l30 * (cs%alphas_muR/twopi)**3
      end if

      if (.not.cs%analytic_expansion) then
         call numerical_expansion(histo_index, v, order, lumi, luminll, luminnll, resummed_expanded)
      else
         call analytic_expansion(histo_index, v, order, lumi, luminll, luminnll, resummed_expanded)
      end if

      ! Hoppet returns x*f(x) - restore correct normalisation
      resummed_expanded(:,:) = resummed_expanded(:,:)/x1/x2
      lumi_asympt(1:) = lumi_asympt(1:)/x1/x2
   end subroutine expanded_sigma

   !-----------------------------------------------------------------------
   subroutine expanded_sigma_inclusive(histo_index, v, order, resummed_expanded, lumi_asympt)
      real(dp),                  intent(in) :: v(:)
      integer,                   intent(in) :: histo_index, order !(0=LL,1=NLL,2=NNLL)
      ! first index of res denotes the order in alphas
      real(dp), intent(out) :: resummed_expanded(0:3,size(v))
      real(dp), optional, intent(out) :: lumi_asympt(0:3)
      !-------------------------------------
      type(LumiExpansion)      :: lumie, luminlle, luminnlle
      type(LumiExpansion_at_x) :: lumi, luminll, luminnll

      if (order < order_NLL) return ! use standard resummation for LL

      call AllocLumiExpansion(lumie)
      if (order > order_NLL) call AllocLumiExpansion(luminlle)

      !--- the following call returns luminosities factors lij, where lij is
      !--- the coefficient of (as2pi(muR))^i * t^j with  t = 2*log(muF1/muF0)=-2*L
      call inclusive_set_lumi_expansion_ptB(pdf1e,pdf2e,lumie, order)
      !--- evaluate luminosity at x=M^2/rts^2
      call inclusive_set_lumi_expansion_at_x(lumi,lumie)

      if (order > order_NLL) then
         call inclusive_set_lumi_expansion_ptB(pdf1e,pdf2e,luminlle,1)
         !--- evaluate luminosity at x=M^2/rts^2
         call inclusive_set_lumi_expansion_at_x(luminll,luminlle)
      end if

      if (order > order_NNLL) call AllocLumiExpansion(luminnlle)

      if (order > order_NNLL) then
         call inclusive_set_lumi_expansion_ptB(pdf1e,pdf2e,luminlle, 1)
         call inclusive_set_lumi_expansion_at_x(luminll,luminlle)
         call inclusive_set_lumi_expansion_ptB(pdf1e,pdf2e,luminnlle, 2)
         call inclusive_set_lumi_expansion_at_x(luminnll,luminnlle)
      end if

      if (present(lumi_asympt)) then
         lumi_asympt = zero ! asymptotic limit for the multiplicative matching
         ! For multiplicative matching scheme, return asymptotic normalization
         if (order == order_NLL)  lumi_asympt(0)  = inclusive_lumi_NLL_loc(cs%muF, zero)
         if (order == order_NNLL) lumi_asympt(0)  = inclusive_lumi_NNLL_loc(cs%muF, zero)
         if (order == order_NNNLL) lumi_asympt(0) = inclusive_lumi_NNNLL_loc(cs%muF, zero)
         ! For multiplicative matching scheme, return the expanded counterterm up to O(as^3). Asymptotic limit goes like:
         lumi_asympt(1) = lumi%l10 * (cs%alphas_muR/twopi)
         lumi_asympt(2) = lumi%l20 * (cs%alphas_muR/twopi)**2
         lumi_asympt(3) = lumi%l30 * (cs%alphas_muR/twopi)**3
      end if

      if (.not.cs%analytic_expansion) then
         call numerical_expansion(histo_index, v, order, lumi, luminll, luminnll, resummed_expanded)
      else
         call analytic_expansion(histo_index, v, order, lumi, luminll, luminnll, resummed_expanded)
      end if

      call Delete(lumie)
      if (order > order_NLL) call Delete(luminlle)
      if (order > order_NNLL) call Delete(luminnlle)

   end subroutine expanded_sigma_inclusive

   ! The following routine computes the expansion as a function of the dimension full quantities v=pt, phi*M=aT,...
   ! corresponding to a kt-ordered evolution.
   !-----------------------------------------------------------------------
   subroutine numerical_expansion(histo_index, v, order, lumi, luminll, luminnll, resummed_expanded)
      real(dp),                  intent(in) :: v(:)
      integer,                   intent(in) :: histo_index, order !(0=LL,1=NLL,2=NNLL)
      ! first index of res denotes the order in alphas
      real(dp), intent(out) :: resummed_expanded(0:3,size(v))
      !-------------------------------------
      type(LumiExpansion_at_x) :: lumi, luminll, luminnll
      real(dp) :: t_array(size(v)), L(size(v))
      logical :: equal
      resummed_expanded = zero

      if (cs%observable=='phi') then
         t_array = v/(cs%Q/cs%M)
      elseif (cs%observable=='pt2') then
         L = Ltilde(two*v/cs%Q, cs%p)
         t_array = two*v/cs%Q
      elseif (cs%observable=="JV_ptB" .or. cs%observable=="JV_ptj") then
         L = Ltilde(min(v/cs%Q, cs%jet_veto/cs%Q), cs%p)
         t_array = v/cs%Q
      else
         L = Ltilde(v/cs%Q, cs%p)
         t_array = v/cs%Q
      end if

      ! if(expansion_call) then
      !   allocate(temp_array(size(v)))
      !   temp_array(:) = t_array(:)
      !   call evaluate_master_integrals(histo_index, t_array)
      !   expansion_call = .false.
      ! else
      !   equal = size(temp_array) == size(t_array)
      !   if(equal) then
      !     if(.not.all(temp_array.eq.t_array)) then
      !       deallocate(temp_array)
      !       allocate(temp_array(size(v)))
      !       temp_array(:) = t_array(:)
      call evaluate_master_integrals(histo_index, t_array)
      !     endif
      !   endif
      ! endif

      ! Born cross section
      resummed_expanded(0,:) = lumi%l00

      ! Higher order corrections in the logarithmic regime
      if (order >= order_NLL) then
         ! perform NLL expansion
         ! O(as)
         resummed_expanded(1,:) = lumi%l10 - 2*lumi%l11*Integrals(histo_index)%MI1(1,:) + 2*cs%ln_Q2_M2*lumi%l00*A(1)*Integrals(histo_index)%MI1(1,:) - 2*lumi%l00*B(1)*Integrals(histo_index)%MI1(1,:) - 2*lumi%l00*A(1)*Integrals(histo_index)%MI1(2,:)

         ! O(as^2)
         resummed_expanded(2,:) = lumi%l20 - 2*lumi%l21*Integrals(histo_index)%MI1(1,:) + 2*cs%ln_Q2_M2*lumi%l10*A(1)*Integrals(histo_index)%MI1(1,:) - 2*lumi%l10*B(1)*Integrals(histo_index)%MI1(1,:) + 4*lumi%l22*Integrals(histo_index)%MI1(2,:) - 2*lumi%l10*A(1)*Integrals(histo_index)%MI1(2,:) - 4*cs%ln_Q2_M2*lumi%l11*A(1)*Integrals(histo_index)%MI1(2,:) + 4*cs%ln_Q2_M2*lumi%l00*beta0*pi*A(1)*Integrals(histo_index)%MI1(2,:) + 4*cs%ln_Q2_muR2*lumi%l00*beta0*pi*A(1)*Integrals(histo_index)%MI1(2,:) + 2*cs%ln_Q2_M2**2*lumi%l00*A(1)**2*Integrals(histo_index)%MI1(2,:) - 2*lumi%l00*A(2)*Integrals(histo_index)%MI1(2,:) + 4*lumi%l11*B(1)*Integrals(histo_index)%MI1(2,:) - 4*lumi%l00*beta0*pi*B(1)*Integrals(histo_index)%MI1(2,:) - 4*cs%ln_Q2_M2*lumi%l00*A(1)*B(1)*Integrals(histo_index)%MI1(2,:) + 2*lumi%l00*B(1)**2*Integrals(histo_index)%MI1(2,:) + 4*lumi%l11*A(1)*Integrals(histo_index)%MI1(3,:) - (16*lumi%l00*beta0*pi*A(1)*Integrals(histo_index)%MI1(3,:))/3. - 4*cs%ln_Q2_M2*lumi%l00*A(1)**2*Integrals(histo_index)%MI1(3,:) + 4*lumi%l00*A(1)*B(1)*Integrals(histo_index)%MI1(3,:) + 2*lumi%l00*A(1)**2*Integrals(histo_index)%MI1(4,:) + 8*lumi%l11*A(1)*Integrals(histo_index)%MI2(1,:) - 8*cs%ln_Q2_M2*lumi%l00*A(1)**2*Integrals(histo_index)%MI2(1,:) + 8*lumi%l00*A(1)*B(1)*Integrals(histo_index)%MI2(1,:) + 16*lumi%l00*A(1)**2*Integrals(histo_index)%MI2(2,:)

         ! O(as^3)
         resummed_expanded(3,:) = lumi%l30 - 2*lumi%l31*Integrals(histo_index)%MI1(1,:) + 2*cs%ln_Q2_M2*lumi%l20*A(1)*Integrals(histo_index)%MI1(1,:) - 2*lumi%l20*B(1)*Integrals(histo_index)%MI1(1,:) + 4*lumi%l32*Integrals(histo_index)%MI1(2,:) - 2*lumi%l20*A(1)*Integrals(histo_index)%MI1(2,:) - 4*cs%ln_Q2_M2*lumi%l21*A(1)*Integrals(histo_index)%MI1(2,:) + 4*cs%ln_Q2_M2*lumi%l10*beta0*pi*A(1)*Integrals(histo_index)%MI1(2,:) + 4*cs%ln_Q2_muR2*lumi%l10*beta0*pi*A(1)*Integrals(histo_index)%MI1(2,:) + 2*cs%ln_Q2_M2**2*lumi%l10*A(1)**2*Integrals(histo_index)%MI1(2,:) - 2*lumi%l10*A(2)*Integrals(histo_index)%MI1(2,:) + 4*lumi%l21*B(1)*Integrals(histo_index)%MI1(2,:) - 4*lumi%l10*beta0*pi*B(1)*Integrals(histo_index)%MI1(2,:) - 4*cs%ln_Q2_M2*lumi%l10*A(1)*B(1)*Integrals(histo_index)%MI1(2,:) + 2*lumi%l10*B(1)**2*Integrals(histo_index)%MI1(2,:) - 8*lumi%l33*Integrals(histo_index)%MI1(3,:) + 4*lumi%l21*A(1)*Integrals(histo_index)%MI1(3,:) + 8*cs%ln_Q2_M2*lumi%l22*A(1)*Integrals(histo_index)%MI1(3,:) - (16*lumi%l10*beta0*pi*A(1)*Integrals(histo_index)%MI1(3,:))/3. - 8*cs%ln_Q2_M2*lumi%l11*beta0*pi*A(1)*Integrals(histo_index)%MI1(3,:) - 8*cs%ln_Q2_muR2*lumi%l11*beta0*pi*A(1)*Integrals(histo_index)%MI1(3,:) + (32*cs%ln_Q2_M2*lumi%l00*beta0**2*pi**2*A(1)*Integrals(histo_index)%MI1(3,:))/3. + (64*cs%ln_Q2_muR2*lumi%l00*beta0**2*pi**2*A(1)*Integrals(histo_index)%MI1(3,:))/3. - (32*lumi%l00*beta1*pi**2*A(1)*Integrals(histo_index)%MI1(3,:))/3. - 4*cs%ln_Q2_M2*lumi%l10*A(1)**2*Integrals(histo_index)%MI1(3,:) - 4*cs%ln_Q2_M2**2*lumi%l11*A(1)**2*Integrals(histo_index)%MI1(3,:) + 8*cs%ln_Q2_M2**2*lumi%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI1(3,:) + 8*cs%ln_Q2_M2*cs%ln_Q2_muR2*lumi%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI1(3,:) + (4*cs%ln_Q2_M2**3*lumi%l00*A(1)**3*Integrals(histo_index)%MI1(3,:))/3. + 4*lumi%l11*A(2)*Integrals(histo_index)%MI1(3,:) - (32*lumi%l00*beta0*pi*A(2)*Integrals(histo_index)%MI1(3,:))/3. - 4*cs%ln_Q2_M2*lumi%l00*A(1)*A(2)*Integrals(histo_index)%MI1(3,:) - 8*lumi%l22*B(1)*Integrals(histo_index)%MI1(3,:) + 8*lumi%l11*beta0*pi*B(1)*Integrals(histo_index)%MI1(3,:) - (32*lumi%l00*beta0**2*pi**2*B(1)*Integrals(histo_index)%MI1(3,:))/3. + 4*lumi%l10*A(1)*B(1)*Integrals(histo_index)%MI1(3,:) + 8*cs%ln_Q2_M2*lumi%l11*A(1)*B(1)*Integrals(histo_index)%MI1(3,:) - 16*cs%ln_Q2_M2*lumi%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI1(3,:) - 8*cs%ln_Q2_muR2*lumi%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI1(3,:) - 4*cs%ln_Q2_M2**2*lumi%l00*A(1)**2*B(1)*Integrals(histo_index)%MI1(3,:) + 4*lumi%l00*A(2)*B(1)*Integrals(histo_index)%MI1(3,:) - 4*lumi%l11*B(1)**2*Integrals(histo_index)%MI1(3,:) + 8*lumi%l00*beta0*pi*B(1)**2*Integrals(histo_index)%MI1(3,:) + 4*cs%ln_Q2_M2*lumi%l00*A(1)*B(1)**2*Integrals(histo_index)%MI1(3,:) - (4*lumi%l00*B(1)**3*Integrals(histo_index)%MI1(3,:))/3. - 8*lumi%l22*A(1)*Integrals(histo_index)%MI1(4,:) + (32*lumi%l11*beta0*pi*A(1)*Integrals(histo_index)%MI1(4,:))/3. - 16*lumi%l00*beta0**2*pi**2*A(1)*Integrals(histo_index)%MI1(4,:) + 2*lumi%l10*A(1)**2*Integrals(histo_index)%MI1(4,:) + 8*cs%ln_Q2_M2*lumi%l11*A(1)**2*Integrals(histo_index)%MI1(4,:) - (56*cs%ln_Q2_M2*lumi%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI1(4,:))/3. - 8*cs%ln_Q2_muR2*lumi%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI1(4,:) - 4*cs%ln_Q2_M2**2*lumi%l00*A(1)**3*Integrals(histo_index)%MI1(4,:) + 4*lumi%l00*A(1)*A(2)*Integrals(histo_index)%MI1(4,:) - 8*lumi%l11*A(1)*B(1)*Integrals(histo_index)%MI1(4,:) + (56*lumi%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI1(4,:))/3. + 8*cs%ln_Q2_M2*lumi%l00*A(1)**2*B(1)*Integrals(histo_index)%MI1(4,:) - 4*lumi%l00*A(1)*B(1)**2*Integrals(histo_index)%MI1(4,:) - 4*lumi%l11*A(1)**2*Integrals(histo_index)%MI1(5,:) + (32*lumi%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI1(5,:))/3. + 4*cs%ln_Q2_M2*lumi%l00*A(1)**3*Integrals(histo_index)%MI1(5,:) - 4*lumi%l00*A(1)**2*B(1)*Integrals(histo_index)%MI1(5,:) - (4*lumi%l00*A(1)**3*Integrals(histo_index)%MI1(6,:))/3. + 8*lumi%l21*A(1)*Integrals(histo_index)%MI2(1,:) - 8*cs%ln_Q2_M2*lumi%l10*A(1)**2*Integrals(histo_index)%MI2(1,:) + 8*lumi%l10*A(1)*B(1)*Integrals(histo_index)%MI2(1,:) - 32*lumi%l22*A(1)*Integrals(histo_index)%MI2(2,:) + 32*lumi%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2(2,:) + 16*lumi%l10*A(1)**2*Integrals(histo_index)%MI2(2,:) + 32*cs%ln_Q2_M2*lumi%l11*A(1)**2*Integrals(histo_index)%MI2(2,:) - 64*cs%ln_Q2_M2*lumi%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(2,:) - 32*cs%ln_Q2_muR2*lumi%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(2,:) - 16*cs%ln_Q2_M2**2*lumi%l00*A(1)**3*Integrals(histo_index)%MI2(2,:) + 16*lumi%l00*A(1)*A(2)*Integrals(histo_index)%MI2(2,:) - 32*lumi%l11*A(1)*B(1)*Integrals(histo_index)%MI2(2,:) + 64*lumi%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2(2,:) + 32*cs%ln_Q2_M2*lumi%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2(2,:) - 16*lumi%l00*A(1)*B(1)**2*Integrals(histo_index)%MI2(2,:) - 48*lumi%l11*A(1)**2*Integrals(histo_index)%MI2(3,:) + 128*lumi%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(3,:) + 48*cs%ln_Q2_M2*lumi%l00*A(1)**3*Integrals(histo_index)%MI2(3,:) - 48*lumi%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2(3,:) - 32*lumi%l00*A(1)**3*Integrals(histo_index)%MI2(4,:) + 16*lumi%l11*A(1)**2*Integrals(histo_index)%MI3(2,:) - 16*cs%ln_Q2_M2*lumi%l00*A(1)**3*Integrals(histo_index)%MI3(2,:) + 16*lumi%l00*A(1)**2*B(1)*Integrals(histo_index)%MI3(2,:) + 32*lumi%l00*A(1)**3*Integrals(histo_index)%MI3(3,:)

      end if

      if (order > order_NLL) then
         ! add NNLL corrections to the above expansion coefficients
         ! no changes at O(as) (virtual corrections are included in the luminosity)

         ! O(as^2)
         resummed_expanded(2,:) = resummed_expanded(2,:) - 2*lumi%l00*(-(cs%ln_Q2_M2*(-2*cs%ln_Q2_muR2*beta0*pi*A(1) + A(2))) - 2*cs%ln_Q2_muR2*beta0*pi*B(1) + B(2))*Integrals(histo_index)%MI1(1,:) + 8*A(1)*((luminll%l11 - cs%ln_Q2_M2*luminll%l00*A(1) + luminll%l00*B(1))*Integrals(histo_index)%MI2(1,:) + 2*luminll%l00*A(1)*Integrals(histo_index)%MI2R(1,:))

         ! additional terms for joint resummation
         if (cs%observable=="JV_ptB" .or. cs%observable=="JV_ptj") then
            ! nb Casimir = A(1)/2
            resummed_expanded(2,:) = resummed_expanded(2,:) +  32._dp * luminll%l00 &
                 & * ( CC**2 * Integrals(histo_index)%MI2clust(1,:) & ! clustering correction
                 ! in units of [dkt/kt*dphi/twopi*dDy*e^{-R}*Lumi]
                 ! clus = 2 (not identical) * 1/2 (rapidity jacobian) * 2 * CC * as/pi * Rp
                 !      = CC * as/pi * 8 * CC * as/pi * L
                 !      = CC^2 * (as/twopi)^2 * 32 * L
                 & + CC * half * ( ca_def/two * Integrals(histo_index)%MI2correlgg(1,:) + nf/two * Integrals(histo_index)%MI2correlqq(1,:)) ) !correlated correction
            ! the factor of 1/2 before the correlated is due to the definition of
            ! the eikonal factor used to compute the MI for the
            ! joint resummation (see double_differential.pdf notes for details)
            !
            ! in units of [dkt/kt*dphi/twopi*dDy*e^{-R}*Lumi]
            ! & using Cnote = Cletter
            ! we need to multiply the integrand of the clustering correction by
            ! 1/2!*Mtilde^2(ka,kb)/(1/2! M^2(ka) * M^2(kb)) = Cletter = Cmaster/(4*CC)
            ! We therefore get:
            ! corr = (CC^2 * (as/twopi)^2 * 32 * L) (clustering) * Cmaster/4/CC
            !      = Cmaster * CC   * (as/twopi)^2 * 8  * L
            !      = Cletter * CC^2 * (as/twopi)^2 * 32 * L
            !      = Cnote   * CC^2 * (as/twopi)^2 * 16 * L
         endif

         ! O(as^3)
         resummed_expanded(3,:) = resummed_expanded(3,:) - 2*(lumi%l10*(-(cs%ln_Q2_M2*(-2*cs%ln_Q2_muR2*beta0*pi*A(1) + A(2))) - 2*cs%ln_Q2_muR2*beta0*pi*B(1) + B(2))*Integrals(histo_index)%MI1(1,:) + (4*cs%ln_Q2_muR2**2*lumi%l00*beta0**2*pi**2*A(1) + 2*cs%ln_Q2_M2**2*lumi%l00*A(1)*(2*cs%ln_Q2_muR2*beta0*pi*A(1) - A(2)) + lumi%l00*A(3) + 4*lumi%l00*beta1*pi**2*B(1) - 4*cs%ln_Q2_muR2*pi*(lumi%l00*beta1*pi*A(1) + lumi%l00*beta0*A(2) + beta0*(-lumi%l11 + 2*lumi%l00*beta0*pi)*B(1) - lumi%l00*beta0*B(1)**2) - 2*lumi%l11*B(2) + 4*lumi%l00*beta0*pi*B(2) - 2*lumi%l00*B(1)*B(2) + 2*cs%ln_Q2_M2*(A(2)*(lumi%l11 - 2*lumi%l00*beta0*pi + lumi%l00*B(1)) - 2*cs%ln_Q2_muR2*beta0*pi*A(1)*(lumi%l11 - 2*lumi%l00*beta0*pi + 2*lumi%l00*B(1)) + lumi%l00*A(1)*(-2*beta1*pi**2 + B(2))))*Integrals(histo_index)%MI1(2,:) + 2*A(1)*(-(lumi%l00*(-(cs%ln_Q2_M2*(-2*cs%ln_Q2_muR2*beta0*pi*A(1) + A(2))) - 2*cs%ln_Q2_muR2*beta0*pi*B(1) + B(2))*Integrals(histo_index)%MI1(3,:)) - 2*(luminll%l21 - cs%ln_Q2_M2*(luminll%l10*A(1) - 2*cs%ln_Q2_muR2*lumi%l00*beta0*pi*A(1) + lumi%l00*A(2)) + luminll%l10*B(1) - 2*cs%ln_Q2_muR2*lumi%l00*beta0*pi*B(1) + lumi%l00*B(2))*Integrals(histo_index)%MI2(1,:) + 4*((2*luminll%l22 - 2*luminll%l11*beta0*pi + 2*cs%ln_Q2_muR2*luminll%l00*beta0*pi*A(1) + cs%ln_Q2_M2**2*luminll%l00*A(1)**2 - luminll%l00*A(2) + 2*luminll%l11*B(1) - 4*luminll%l00*beta0*pi*B(1) + luminll%l00*B(1)**2 - 2*cs%ln_Q2_M2*A(1)*(luminll%l11 - 2*luminll%l00*beta0*pi + luminll%l00*B(1)))*Integrals(histo_index)%MI2(2,:) + A(1)*((luminll%l11 - cs%ln_Q2_M2*luminll%l00*A(1) + luminll%l00*B(1))*Integrals(histo_index)%MI2(3,:) - luminll%l10*Integrals(histo_index)%MI2R(1,:) + 2*(luminll%l11 - 6*luminll%l00*beta0*pi - cs%ln_Q2_M2*luminll%l00*A(1) + luminll%l00*B(1))*Integrals(histo_index)%MI2R(2,:) + 2*luminll%l00*A(1)*Integrals(histo_index)%MI2R(3,:) - 2*luminll%l11*Integrals(histo_index)%MI3(2,:) + 2*cs%ln_Q2_M2*luminll%l00*A(1)*Integrals(histo_index)%MI3(2,:) - 2*luminll%l00*B(1)*Integrals(histo_index)%MI3(2,:) - 2*luminll%l00*A(1)*Integrals(histo_index)%MI3R(2,:)))))
      end if

      ! add N3LL corrections for ptB
      if (order > order_NNLL) then
         resummed_expanded(2,:) = resummed_expanded(2,:)  + (lumi%p0p020 + 4*cs%ln_Q2_M2**2*luminnll%l00*A(1)**2 + 8*luminnll%l11*B(1) + 4*luminnll%l00*B(1)**2 - 8*cs%ln_Q2_M2*A(1)*(luminnll%l11 + luminnll%l00*B(1)))*Integrals(histo_index)%MI2(0,:) + 8*A(1)*((-luminll%l11 + luminnll%l11 - cs%ln_Q2_M2*(-luminll%l00 + luminnll%l00)*A(1) - luminll%l00*B(1) + luminnll%l00*B(1))*Integrals(histo_index)%MI2(1,:) + (luminnll%l11 - cs%ln_Q2_M2*luminnll%l00*A(1) + luminnll%l00*B(1))*Integrals(histo_index)%MI2R(0,:) + 2*(-luminll%l00 + luminnll%l00)*A(1)*Integrals(histo_index)%MI2R(1,:))

         resummed_expanded(3,:) = resummed_expanded(3,:) - 2*(2*(cs%ln_Q2_M2**2*A(1)*(-(luminnll%l10*A(1)) + 4*cs%ln_Q2_muR2*luminnll%l00*beta0*pi*A(1) - 2*luminnll%l00*A(2)) - 2*luminnll%l21*B(1) - luminnll%l10*B(1)**2 + 4*cs%ln_Q2_muR2*beta0*pi*B(1)*(luminnll%l11 + luminnll%l00*B(1)) - 2*luminnll%l11*B(2) - 2*luminnll%l00*B(1)*B(2) + 2*cs%ln_Q2_M2*(A(2)*(luminnll%l11 + luminnll%l00*B(1)) - 2*cs%ln_Q2_muR2*beta0*pi*A(1)*(luminnll%l11 + 2*luminnll%l00*B(1)) + A(1)*(luminnll%l21 + luminnll%l10*B(1) + luminnll%l00*B(2))))*Integrals(histo_index)%MI2(0,:) + lumi%p0p031*Integrals(histo_index)%MI2(1,:) + 4*luminll%l21*A(1)*Integrals(histo_index)%MI2(1,:) - 4*luminnll%l21*A(1)*Integrals(histo_index)%MI2(1,:) - 16*cs%ln_Q2_M2*luminnll%l22*A(1)*Integrals(histo_index)%MI2(1,:) - cs%ln_Q2_M2*lumi%p0p020*A(1)*Integrals(histo_index)%MI2(1,:) + 16*cs%ln_Q2_M2*luminnll%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2(1,:) + 16*cs%ln_Q2_muR2*luminnll%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2(1,:) - 4*cs%ln_Q2_M2*luminll%l10*A(1)**2*Integrals(histo_index)%MI2(1,:) + 4*cs%ln_Q2_M2*luminnll%l10*A(1)**2*Integrals(histo_index)%MI2(1,:) + 12*cs%ln_Q2_M2**2*luminnll%l11*A(1)**2*Integrals(histo_index)%MI2(1,:) - 16*cs%ln_Q2_M2**2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(1,:) - 24*cs%ln_Q2_M2*cs%ln_Q2_muR2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(1,:) - 4*cs%ln_Q2_M2**3*luminnll%l00*A(1)**3*Integrals(histo_index)%MI2(1,:) - 8*luminnll%l11*A(2)*Integrals(histo_index)%MI2(1,:) + 12*cs%ln_Q2_M2*luminnll%l00*A(1)*A(2)*Integrals(histo_index)%MI2(1,:) + 16*luminnll%l22*B(1)*Integrals(histo_index)%MI2(1,:) + lumi%p0p020*B(1)*Integrals(histo_index)%MI2(1,:) - 16*luminnll%l11*beta0*pi*B(1)*Integrals(histo_index)%MI2(1,:) + 4*luminll%l10*A(1)*B(1)*Integrals(histo_index)%MI2(1,:) - 4*luminnll%l10*A(1)*B(1)*Integrals(histo_index)%MI2(1,:) - 24*cs%ln_Q2_M2*luminnll%l11*A(1)*B(1)*Integrals(histo_index)%MI2(1,:) + 32*cs%ln_Q2_M2*luminnll%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2(1,:) + 24*cs%ln_Q2_muR2*luminnll%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2(1,:) + 12*cs%ln_Q2_M2**2*luminnll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2(1,:) - 8*luminnll%l00*A(2)*B(1)*Integrals(histo_index)%MI2(1,:) + 12*luminnll%l11*B(1)**2*Integrals(histo_index)%MI2(1,:) - 16*luminnll%l00*beta0*pi*B(1)**2*Integrals(histo_index)%MI2(1,:) - 12*cs%ln_Q2_M2*luminnll%l00*A(1)*B(1)**2*Integrals(histo_index)%MI2(1,:) + 4*luminnll%l00*B(1)**3*Integrals(histo_index)%MI2(1,:) - 4*luminnll%l00*A(1)*B(2)*Integrals(histo_index)%MI2(1,:) - 16*luminll%l22*A(1)*Integrals(histo_index)%MI2(2,:) + 16*luminnll%l22*A(1)*Integrals(histo_index)%MI2(2,:) + lumi%p0p020*A(1)*Integrals(histo_index)%MI2(2,:) + 16*luminll%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2(2,:) - 16*luminnll%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2(2,:) + 16*cs%ln_Q2_M2*luminll%l11*A(1)**2*Integrals(histo_index)%MI2(2,:) - 24*cs%ln_Q2_M2*luminnll%l11*A(1)**2*Integrals(histo_index)%MI2(2,:) - 32*cs%ln_Q2_M2*luminll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(2,:) - 16*cs%ln_Q2_muR2*luminll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(2,:) + 32*cs%ln_Q2_M2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(2,:) + 16*cs%ln_Q2_muR2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2(2,:) - 8*cs%ln_Q2_M2**2*luminll%l00*A(1)**3*Integrals(histo_index)%MI2(2,:) + 12*cs%ln_Q2_M2**2*luminnll%l00*A(1)**3*Integrals(histo_index)%MI2(2,:) + 8*luminll%l00*A(1)*A(2)*Integrals(histo_index)%MI2(2,:) - 8*luminnll%l00*A(1)*A(2)*Integrals(histo_index)%MI2(2,:) - 16*luminll%l11*A(1)*B(1)*Integrals(histo_index)%MI2(2,:) + 24*luminnll%l11*A(1)*B(1)*Integrals(histo_index)%MI2(2,:) + 32*luminll%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2(2,:) - 32*luminnll%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2(2,:) + 16*cs%ln_Q2_M2*luminll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2(2,:) - 24*cs%ln_Q2_M2*luminnll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2(2,:) - 8*luminll%l00*A(1)*B(1)**2*Integrals(histo_index)%MI2(2,:) + 12*luminnll%l00*A(1)*B(1)**2*Integrals(histo_index)%MI2(2,:) - 8*luminll%l11*A(1)**2*Integrals(histo_index)%MI2(3,:) + 8*luminnll%l11*A(1)**2*Integrals(histo_index)%MI2(3,:) + 8*cs%ln_Q2_M2*luminll%l00*A(1)**3*Integrals(histo_index)%MI2(3,:) - 8*cs%ln_Q2_M2*luminnll%l00*A(1)**3*Integrals(histo_index)%MI2(3,:) - 8*luminll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2(3,:) + 8*luminnll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2(3,:) - 4*luminnll%l21*A(1)*Integrals(histo_index)%MI2R(0,:) + cs%ln_Q2_M2*lumi%p020*A(1)*Integrals(histo_index)%MI2R(0,:) + 8*cs%ln_Q2_M2*luminnll%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2R(0,:) + 8*cs%ln_Q2_muR2*luminnll%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2R(0,:) + 4*cs%ln_Q2_M2*luminnll%l10*A(1)**2*Integrals(histo_index)%MI2R(0,:) - 8*cs%ln_Q2_M2**2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2R(0,:) - 16*cs%ln_Q2_M2*cs%ln_Q2_muR2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2R(0,:) - 4*luminnll%l11*A(2)*Integrals(histo_index)%MI2R(0,:) + 8*cs%ln_Q2_M2*luminnll%l00*A(1)*A(2)*Integrals(histo_index)%MI2R(0,:) - lumi%p020*B(1)*Integrals(histo_index)%MI2R(0,:) - 8*luminnll%l11*beta0*pi*B(1)*Integrals(histo_index)%MI2R(0,:) - 4*luminnll%l10*A(1)*B(1)*Integrals(histo_index)%MI2R(0,:) + 16*cs%ln_Q2_M2*luminnll%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2R(0,:) + 16*cs%ln_Q2_muR2*luminnll%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2R(0,:) - 4*luminnll%l00*A(2)*B(1)*Integrals(histo_index)%MI2R(0,:) - 8*luminnll%l00*beta0*pi*B(1)**2*Integrals(histo_index)%MI2R(0,:) - 4*luminnll%l00*A(1)*B(2)*Integrals(histo_index)%MI2R(0,:) + 16*luminnll%l22*A(1)*Integrals(histo_index)%MI2R(1,:) - 2*lumi%p020*A(1)*Integrals(histo_index)%MI2R(1,:) - 32*luminnll%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2R(1,:) + 8*luminll%l10*A(1)**2*Integrals(histo_index)%MI2R(1,:) - 8*luminnll%l10*A(1)**2*Integrals(histo_index)%MI2R(1,:) - 16*cs%ln_Q2_M2*luminnll%l11*A(1)**2*Integrals(histo_index)%MI2R(1,:) + 64*cs%ln_Q2_M2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2R(1,:) + 32*cs%ln_Q2_muR2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2R(1,:) + 8*cs%ln_Q2_M2**2*luminnll%l00*A(1)**3*Integrals(histo_index)%MI2R(1,:) - 16*luminnll%l00*A(1)*A(2)*Integrals(histo_index)%MI2R(1,:) + 16*luminnll%l11*A(1)*B(1)*Integrals(histo_index)%MI2R(1,:) - 64*luminnll%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2R(1,:) - 16*cs%ln_Q2_M2*luminnll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2R(1,:) + 8*luminnll%l00*A(1)*B(1)**2*Integrals(histo_index)%MI2R(1,:) - 16*luminll%l11*A(1)**2*Integrals(histo_index)%MI2R(2,:) + 24*luminnll%l11*A(1)**2*Integrals(histo_index)%MI2R(2,:) + 96*luminll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2R(2,:) - 96*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2R(2,:) + 16*cs%ln_Q2_M2*luminll%l00*A(1)**3*Integrals(histo_index)%MI2R(2,:) - 24*cs%ln_Q2_M2*luminnll%l00*A(1)**3*Integrals(histo_index)%MI2R(2,:) - 16*luminll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2R(2,:) + 24*luminnll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI2R(2,:) - 16*luminll%l00*A(1)**3*Integrals(histo_index)%MI2R(3,:) + 16*luminnll%l00*A(1)**3*Integrals(histo_index)%MI2R(3,:) - 16*luminnll%l11*beta0*pi*A(1)*Integrals(histo_index)%MI2R2(0,:) + 16*cs%ln_Q2_M2*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2R2(0,:) - 16*luminnll%l00*beta0*pi*A(1)*B(1)*Integrals(histo_index)%MI2R2(0,:) - 32*luminnll%l00*beta0*pi*A(1)**2*Integrals(histo_index)%MI2R2(1,:) - 3*lumi%p0p020*A(1)*Integrals(histo_index)%MI3(1,:) + 8*cs%ln_Q2_M2*luminll%l11*A(1)**2*Integrals(histo_index)%MI3(1,:) + 16*cs%ln_Q2_M2*luminnll%l11*A(1)**2*Integrals(histo_index)%MI3(1,:) - 4*cs%ln_Q2_M2**2*luminll%l00*A(1)**3*Integrals(histo_index)%MI3(1,:) - 8*cs%ln_Q2_M2**2*luminnll%l00*A(1)**3*Integrals(histo_index)%MI3(1,:) - 8*luminll%l11*A(1)*B(1)*Integrals(histo_index)%MI3(1,:) - 16*luminnll%l11*A(1)*B(1)*Integrals(histo_index)%MI3(1,:) + 8*cs%ln_Q2_M2*luminll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI3(1,:) + 16*cs%ln_Q2_M2*luminnll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI3(1,:) - 4*luminll%l00*A(1)*B(1)**2*Integrals(histo_index)%MI3(1,:) - 8*luminnll%l00*A(1)*B(1)**2*Integrals(histo_index)%MI3(1,:) + 16*luminll%l11*A(1)**2*Integrals(histo_index)%MI3(2,:) - 16*luminnll%l11*A(1)**2*Integrals(histo_index)%MI3(2,:) - 16*cs%ln_Q2_M2*luminll%l00*A(1)**3*Integrals(histo_index)%MI3(2,:) + 16*cs%ln_Q2_M2*luminnll%l00*A(1)**3*Integrals(histo_index)%MI3(2,:) + 16*luminll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI3(2,:) - 16*luminnll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI3(2,:) - 8*luminll%l11*A(1)**2*Integrals(histo_index)%MI3R(1,:) - 8*luminnll%l11*A(1)**2*Integrals(histo_index)%MI3R(1,:) + 8*cs%ln_Q2_M2*luminll%l00*A(1)**3*Integrals(histo_index)%MI3R(1,:) + 8*cs%ln_Q2_M2*luminnll%l00*A(1)**3*Integrals(histo_index)%MI3R(1,:) - 8*luminll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI3R(1,:) - 8*luminnll%l00*A(1)**2*B(1)*Integrals(histo_index)%MI3R(1,:) + 16*luminll%l00*A(1)**3*Integrals(histo_index)%MI3R(2,:) - 16*luminnll%l00*A(1)**3*Integrals(histo_index)%MI3R(2,:) - 16*luminll%l00*A(1)**3*Integrals(histo_index)%MI3RR(1,:) + lumi%l00*Integrals(histo_index)%MI1(1,:)*(-(cs%ln_Q2_M2*(4*cs%ln_Q2_muR2**2*beta0**2*pi**2*A(1) - 4*cs%ln_Q2_muR2*pi*(beta1*pi*A(1) + beta0*A(2)) + A(3))) + 4*cs%ln_Q2_muR2**2*beta0**2*pi**2*B(1) - 4*cs%ln_Q2_muR2*pi*(beta1*pi*B(1) + beta0*B(2)) + B(3) - 16*beta0**2*pi**2*A(1)*zeta3))
      end if

      resummed_expanded(1,:) = resummed_expanded(1,:) * (cs%alphas_muR/twopi)
      resummed_expanded(2,:) = resummed_expanded(2,:) * (cs%alphas_muR/twopi)**2
      resummed_expanded(3,:) = resummed_expanded(3,:) * (cs%alphas_muR/twopi)**3

   end subroutine numerical_expansion

   ! This routines computes the analytical expansion as a function of the dimension full quantities v=pt, phi*M=aT,...
   ! WARNING: this expansion is for internal test only, do not use it unless you know what you are doing
   ! The analytic expansion is valid for central scales only and for NNLL, N3LL for a limited number of observables
   !-------------------------------------------------------
   subroutine analytic_expansion(histo_index, v, order, lumi, luminll, luminnll, resummed_expanded)
      real(dp),                  intent(in) :: v(:)
      integer,                   intent(in) :: histo_index, order !(2=NNLL, 3=N3LL)
      ! first index of res denotes the order in alphas
      real(dp), intent(out) :: resummed_expanded(0:3,size(v))
      !-------------------------------------
      type(LumiExpansion_at_x) :: lumi, luminll, luminnll
      real(dp) :: L(size(v)), lambda(size(v))
      integer  :: i
      real(dp) :: t_array(size(v))
      real(dp), parameter :: zeta4=pi**4/90._dp, zeta5=1.0369277551433699263313654864570
      integer :: mu, delta, gamma
      real(dp) :: h11, h12, h21, h22, h23, h24

      ! for the checks run with (plot log($1):($1*$5)):
      ! -proc H -resum-only -out radish2.test -vmin 1d-1 -vmax 10d0 -nbins 25 -sqrts 13000 -pdf_name PDF4LHC15_nnlo_mc.LHgrid -observable ptB -p 2 -order 3 -fixed_order 3 -differential -ncall1 1 -ncall2 1 -nev 1 -nocuts -xmuR 1.0 -xmuF 1.0 -xQ 1.0  -analytic-expansion
      ! or, for phistar, use
      ! -proc DY -resum-only -out radish2_analytic_right.test -vmin 1d-4 -vmax 1d0 -nbins 50 -log -sqrts 13000 -pdf_name PDF4LHC15_nnlo_mc.LHgrid -observable phi -p 2 -order 3 -fixed_order 3 -differential -ncall1 1 -ncall2 1 -nev 1 -nocuts -xmuR 1.0 -xmuF 1.0 -xQ 1.0 -dynamical -analytic-expansion
      ! TODO: is this comment still valid? WARNING, IMPORTANT: to run this test, remember to comment the term proportional to zeta3 in msqV2 in get_virtual.
      ! Since the scale dependence is not implemented in this routine, the test only works for muR=muF=Q (either static or dynamic)


      if (cs%observable=='phi') then
         L = Ltilde(v/(cs%Q/cs%M), cs%p)
      elseif (cs%observable=='pt2') then
         L = Ltilde(two*v/cs%Q, cs%p)
      elseif (cs%observable=="JV_ptB" .or. cs%observable=="JV_ptj") then
         L = Ltilde(min(v/cs%Q, cs%jet_veto/cs%Q), cs%p)
         t_array = v/cs%Q
         ! master integrals needed for correlated and clustered terms
         call evaluate_master_integrals(histo_index, t_array)
      else
         L = Ltilde(v/cs%Q, cs%p)
      end if
      lambda  = get_lambda(L)

      resummed_expanded = zero

      if (order < order_NLL) return ! use standard resummation for LL

      if (order == order_NLL) then
         call wae_error("This routine is only available at NNLL or at N3LL.")
      endif

      ! NNLL and N3LL expansion including luminosities:
      mu=1
      delta=1
      gamma=1

      if (order == order_NNLL) then

         if (cs%observable == 'ptB' .or. cs%observable=="JV_ptB") then

            resummed_expanded(0,:) = lumi%l00

            resummed_expanded(1,:) = lumi%l10 - 2*L*lumi%l11 - 2*L**2*lumi%l00*A(1) - 2*L*lumi%l00*B(1)

            resummed_expanded(2,:) = lumi%l20 - (4*beta0*L**2*lumi%l00*pi*(4*L*A(1) + 3*B(1)))/3. - 2*gamma*zeta3*A(1)*(lumi%l11 + 2*L*lumi%l00*A(1) + lumi%l00*B(1)) &
            &  + 2*L*(-lumi%l21 + 2*L*lumi%l22 - L*lumi%l00*A(2) + (L*A(1) + B(1))*(-lumi%l10 + L*(2*lumi%l11 + L*lumi%l00*A(1) + lumi%l00*B(1))) &
            &  - lumi%l00*B(2))

            ! additional terms for joint resummation
            if (cs%observable=="JV_ptB") then
               ! nb Casimir = A(1)/2
               resummed_expanded(2,:) = resummed_expanded(2,:) +  32._dp * luminll%l00 &
                  & * ( CC**2 * Integrals(histo_index)%MI2clust(1,:) & ! clustering correction
                  & + CC * half * ( ca_def/two * Integrals(histo_index)%MI2correlgg(1,:) + nf/two * Integrals(histo_index)%MI2correlqq(1,:)) ) !correlated correction
                  ! the factor of 1/2 before the correlated is due to the definition of
                  ! the eikonal factor used to compute the MI for the
                  ! joint resummation (see double_differential.pdf notes for details)
            endif

            if (cs%observable == 'ptB') then
               resummed_expanded(3,:) = (3*lumi%l30 - 6*L**2*lumi%l00*A(3) - 2*L*(L**3*(-3*lumi%l10 + 6*L*lumi%l11 - 16*beta0*L*lumi%l00*pi)*A(1)**2 + 2*L**5*lumi%l00*A(1)**3 + &
               &       3*L*(-lumi%l10 + 2*L*(lumi%l11 - 2*beta0*lumi%l00*pi + L*lumi%l00*A(1)))*B(1)**2 + 2*L**2*lumi%l00*B(1)**3 + &
               &       4*beta0*L*lumi%l00*pi*(4*L*A(2) + 3*B(2)) + L*A(1)* &
               &        (3*lumi%l20 + 2*L*(-3*lumi%l21 + 6*L*lumi%l22 + 4*beta0*L*pi*(-2*lumi%l11 + 3*beta0*lumi%l00*pi) + &
               &             4*pi*(beta0*lumi%l10 + 2*beta1*lumi%l00*pi)) - 6*L**2*lumi%l00*A(2) - 6*L*lumi%l00*B(2)) + &
               &       B(1)*(3*lumi%l20 + 2*L*(-3*lumi%l21 + 6*L*lumi%l22 + 2*beta0*L*pi*(-3*lumi%l11 + 4*beta0*lumi%l00*pi) + &
               &             3*pi*(beta0*lumi%l10 + 2*beta1*lumi%l00*pi) + L*(-3*lumi%l10 + 6*L*lumi%l11 - 14*beta0*L*lumi%l00*pi)*A(1) + &
               &             3*L**3*lumi%l00*A(1)**2) - 6*L**2*lumi%l00*A(2) - 6*L*lumi%l00*B(2)) + &
               &       3*(lumi%l31 + L*(-2*lumi%l32 + 4*L*lumi%l33 + (lumi%l10 - 2*L*lumi%l11)*A(2)) + (lumi%l10 - 2*L*lumi%l11)*B(2))) + &
               &    2*zeta3*((2*L*A(1) + B(1))**2*(3*lumi%l11 + 2*L*lumi%l00*A(1) + lumi%l00*B(1)) - &
               &       3*gamma*(-4*L**3*lumi%l00*A(1)**3 + 2*beta0*pi*B(1)*(lumi%l11 + lumi%l00*B(1)) + A(2)*(lumi%l11 + 4*L*lumi%l00*A(1) + lumi%l00*B(1)) + &
               &          2*L*A(1)**2*(lumi%l10 - 3*L*(lumi%l11 - 4*beta0*lumi%l00*pi + lumi%l00*B(1))) + &
               &          A(1)*(lumi%l21 - 4*L*lumi%l22 + lumi%l10*B(1) - 4*L*lumi%l11*B(1) - 2*L*lumi%l00*B(1)**2 + &
               &             8*beta0*L*pi*(lumi%l11 + 2*lumi%l00*B(1)) + lumi%l00*B(2)))))/3.
            endif

         else if (cs%observable == 'phi') then

            resummed_expanded(0,:) = lumi%l00

            resummed_expanded(1,:) = 0.16666666666666666*(0. + 6.*lumi%l10 + 8.317766166719343*lumi%l11 - 12.*L*lumi%l11 - 15.635040568107774*lumi%l00*A(1) + &
            &    16.635532333438686*L*lumi%l00*A(1) - 12.*L**2*lumi%l00*A(1) + 8.317766166719343*lumi%l00*B(1) - 12.*L*lumi%l00*B(1))

            resummed_expanded(2,:) =      39.47841760435743*(5.346865757648092e-6*(2.*L*A(1) + B(1))*(-84276.83131639716*lumi%l00*A(1) + 12344.933216176572*lumi%l00*(2.*L*A(1) + B(1))) + 0.0001679767418395807*(-2682.614858425291*lumi%l11*A(1) + 785.9028573975323*lumi%l11*(2.*L*A(1) + B(1)) - 0.3183098861837907*(L**2*A(1) + L*B(1))*(-2468.986643235314*lumi%l00*A(1) + 1313.4889850605662*lumi%l00*(2.*L*A(1) + B(1))) + 15.635040568107774*(-25.132741228718345*(lumi%l10 - 2.*L*lumi%l11)*A(1) - 25.132741228718345*lumi%l00*(0. + 25.132741228718345*beta0*L*A(1) + A(2) + 6.283185307179586*beta0*B(1))) + 16.635532333438686*(12.566370614359172*(lumi%l10 - 2.*L*lumi%l11)*(2.*L*A(1) + B(1)) + 12.566370614359172*lumi%l00*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2)))) + 0.004031441804149937*(6.283185307179586*(lumi%l20 + 1.3862943611198906*lumi%l21 + 4.*L**2*lumi%l22 - 2.*L*(lumi%l21 + 2.772588722239781*lumi%l22)) - 12.566370614359172*(lumi%l10 + (1.3862943611198906 - 2.*L)*lumi%l11)*(L**2*A(1) + L*B(1)) + 124.02510672119926*lumi%l00*(0.10132118364233778*(L**2*A(1) + L*B(1))**2 - 0.03377372788077926*(0. + 25.132741228718345*beta0*L**3*A(1) + 3.*L**2*A(2) + 18.84955592153876*beta0*L**2*B(1) + 3.*L*B(2)))));

            resummed_expanded(3,:) =       248.05021344239853*(1.4183001922974766e-7*(2.*L*A(1) + B(1))**2*(-1.930469708385928e6*lumi%l00*A(1) + 168553.66263279432*lumi%l00*(2.*L*A(1) + B(1))) + 0.00001679767418395807*((-84276.83131639716*lumi%l00*A(1) + 12344.933216176572*lumi%l00*(2.*L*A(1) + B(1)))*(-0.10132118364233778*(2.*L*A(1) + B(1))*(L**2*A(1) + L*B(1)) + 0.05066059182116889*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2))) + 0.3183098861837907*(2.*L*A(1) + B(1))*(-102414.59885959674*lumi%l11*A(1) + 13413.074292126454*lumi%l11*(2.*L*A(1) + B(1)) - 533.6892689126876*(25.132741228718345*(lumi%l10 - 2.*L*lumi%l11)*A(1) + 25.132741228718345*lumi%l00*(0. + 25.132741228718345*beta0*L*A(1) + A(2) + 6.283185307179586*beta0*B(1))) + 156.35040568107775*(12.566370614359172*(lumi%l10 - 2.*L*lumi%l11)*(2.*L*A(1) + B(1)) + 12.566370614359172*lumi%l00*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2))))) + 0.0001679767418395807*(-106.73785378253751*(4.*(lumi%l21 - 4.*L*lumi%l22)*A(1) + 4.*lumi%l11*(0. + 25.132741228718345*beta0*L*A(1) + A(2) + 6.283185307179586*beta0*B(1))) + 62.5401622724311*(2.*(lumi%l21 - 4.*L*lumi%l22)*(2.*L*A(1) + B(1)) + 2.*lumi%l11*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2))) + 0.5*(-2468.986643235314*lumi%l00*A(1) + 1313.4889850605662*lumi%l00*(2.*L*A(1) + B(1)))*(0.10132118364233778*(L**2*A(1) + L*B(1))**2 - 0.03377372788077926*(0. + 25.132741228718345*beta0*L**3*A(1) + 3.*L**2*A(2) + 18.84955592153876*beta0*L**2*B(1) + 3.*L*B(2))) + 16.635532333438686*(2.*(lumi%l20 - 2.*L*lumi%l21 + 4.*L**2*lumi%l22)*(2.*L*A(1) + B(1)) + 2.*(lumi%l10 - 2.*L*lumi%l11)*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2)) + 4.*lumi%l00*(0. + 78.95683520871486*beta1*L**2*A(1) + 157.91367041742973*beta0**2*L**3*A(1) + 25.132741228718345*beta0*L**2*A(2) + L*A(3) + 39.47841760435743*beta1*L*B(1) + 78.95683520871486*beta0**2*L**2*B(1) + 12.566370614359172*beta0*L*B(2))) - 0.3183098861837907*(L**2*A(1) + L*B(1))*(-2682.614858425291*lumi%l11*A(1) + 785.9028573975323*lumi%l11*(2.*L*A(1) + B(1)) + 15.635040568107774*(-25.132741228718345*(lumi%l10 - 2.*L*lumi%l11)*A(1) - 25.132741228718345*lumi%l00*(0. + 25.132741228718345*beta0*L*A(1) + A(2) + 6.283185307179586*beta0*B(1))) + 16.635532333438686*(12.566370614359172*(lumi%l10 - 2.*L*lumi%l11)*(2.*L*A(1) + B(1)) + 12.566370614359172*lumi%l00*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2)))) + 15.635040568107774*(-4.*(lumi%l20 - 2.*L*lumi%l21 + 4.*L**2*lumi%l22)*A(1) - 4.*(lumi%l10 - 2.*L*lumi%l11)*(0. + 25.132741228718345*beta0*L*A(1) + A(2) + 6.283185307179586*beta0*B(1)) - 248.05021344239853*lumi%l00*(7.639437268410976*beta0**2*L**2*A(1) + 2.5464790894703255*(0. + beta1*L)*A(1) + 0.8105694691387022*beta0*L*A(2) + 2.5464790894703255*beta0**2*L*B(1) + 0.3183098861837907*(0. + 0.05066059182116889*A(3) - 2.*(0. - 1.*beta1)*B(1) + 0.6366197723675814*beta0*B(2))))) + 0.004031441804149937*(lumi%l30 + 1.3862943611198906*lumi%l31 - 2.*L*(lumi%l31 + 2.772588722239781*lumi%l32) - 8.*L**3*lumi%l33 + 4.*L**2*(lumi%l32 + 4.1588830833596715*lumi%l33) - 2.*(lumi%l20 + 1.3862943611198906*lumi%l21 + 4.*L**2*lumi%l22 - 2.*L*(lumi%l21 + 2.772588722239781*lumi%l22))*(L**2*A(1) + L*B(1)) + 19.739208802178716*(lumi%l10 + (1.3862943611198906 - 2.*L)*lumi%l11)*(0.10132118364233778*(L**2*A(1) + L*B(1))**2 - 0.03377372788077926*(0. + 25.132741228718345*beta0*L**3*A(1) + 3.*L**2*A(2) + 18.84955592153876*beta0*L**2*B(1) + 3.*L*B(2))) + 82.68340448079951*lumi%l00*(0.010750511477733165*(L**2*A(1) + L*B(1))*(0. + 25.132741228718345*beta0*L**3*A(1) + 3.*L**2*A(2) + 18.84955592153876*beta0*L**2*B(1) + 3.*L*B(2)) - 0.15915494309189535*(L**2*A(1) + L*B(1))*(0.10132118364233778*(L**2*A(1) + L*B(1))**2 - 0.03377372788077926*(0. + 25.132741228718345*beta0*L**3*A(1) + 3.*L**2*A(2) + 18.84955592153876*beta0*L**2*B(1) + 3.*L*B(2))) + 3.*(-0.6366197723675814*beta0**2*L**4*A(1) - 0.4244131815783876*(0. + beta1*L**3)*A(1) - 0.13509491152311703*beta0*L**3*A(2) - 0.4244131815783876*beta0**2*L**3*B(1) - 0.3183098861837907*(0. + 0.025330295910584444*L**2*A(3) - 1.*(0. - 1.*beta1*L**2)*B(1) + 0.3183098861837907*beta0*L**2*B(2))))))

         else if (cs%observable == 'JV_ptj') then

            resummed_expanded(0,:) = lumi%l00

            !-- LL terms
            h12 = -two**2*CC
            h11 = zero
            h21 = zero
            h22 = zero
            h23 = -32._dp*CC*Pi*beta0/(3._dp)
            h24 = two**3*CC**2

            !-- NLL terms
            h11 = h11 - 8._dp*BB*CC+4._dp*CC*cs%ln_Q2_M2
            h22 = h22 + 32*BB**2*CC**2-16*BB*CC*beta0*Pi-32*BB*CC**2*cs%ln_Q2_M2 &
               & + 16*CC*beta0*Pi*cs%ln_Q2_M2+8*CC**2*cs%ln_Q2_M2**2-8*CC*beta0*Pi*cs%ln_muR2_M2 &
               & - 4*CC*cmw_K
            h23 = h23 + 32*BB*CC**2-16*CC**2*cs%ln_Q2_M2
            ! NNLL terms
            h21 = h21 - two*B(2)-16*BB*CC*beta0*Pi*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
               & +8*CC*beta0*Pi*cs%ln_Q2_M2*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
               & +4._dp*CC*cmw_K*cs%ln_Q2_M2

            !---------------------------------------------------------------------------------

            ! Build the expansion
            !-- LL terms
            resummed_expanded(1,:) = lumi%l00*h12*L**2
            resummed_expanded(2,:) = lumi%l00*(h24*L**4 + h23*L**3)

            resummed_expanded(1,:) = resummed_expanded(1,:) + (lumi%l00*h11 - two*lumi%l11)*L + lumi%l10
            resummed_expanded(2,:) = resummed_expanded(2,:) + (-two*lumi%l11*h12)*L**3 &
               & + (lumi%l00*h22 + lumi%l10*h12 - two*lumi%l11*h11 + four*lumi%l22)*L**2 &
               & + (lumi%l00*h21 + lumi%l10*h11 - two*lumi%l21)*L  + lumi%l20 &
               & +  32._dp * luminll%l00 &
               & * ( CC**2 * Integrals(histo_index)%MI2clust(1,:) & ! clustering correction
               & + CC * half * ( ca_def/two * Integrals(histo_index)%MI2correlgg(1,:) + nf/two * Integrals(histo_index)%MI2correlqq(1,:)) ) !correlated correction
               ! the factor of 1/2 before the correlated is due to the definition of
               ! the eikonal factor used to compute the MI for the
               ! joint resummation (see double_differential.pdf notes for details)
         endif

      else if (order == order_NNNLL) then

         if (cs%observable == 'ptB') then
            resummed_expanded(0,:) = lumi%l00

            resummed_expanded(1,:) = lumi%l10 - 2*L*lumi%l11 - 2*L**2*lumi%l00*A(1) - 2*L*lumi%l00*B(1)

            resummed_expanded(2,:) = (3*lumi%l20 - 4*beta0*L**2*lumi%l00*pi*(4*L*A(1) + 3*B(1)) - 2*zeta3*A(1)*(-4*beta0*delta*lumi%l00*pi + 3*gamma*(lumi%l11 + 2*L*lumi%l00*A(1) + lumi%l00*B(1))) + 6*L*(-lumi%l21 + 2*L*lumi%l22 - L*lumi%l00*A(2) + (L*A(1) + B(1))*(-lumi%l10 + L*(2*lumi%l11 + L*lumi%l00*A(1) + lumi%l00*B(1))) - lumi%l00*B(2)))/3.

            resummed_expanded(3,:) = (3*lumi%l30 - 6*L**2*lumi%l00*A(3) - 2*L*(L**3*(-3*lumi%l10 + 6*L*lumi%l11 - 16*beta0*L*lumi%l00*pi)*A(1)**2 + 2*L**5*lumi%l00*A(1)**3 + 3*L*(-lumi%l10 + 2*L*(lumi%l11 - 2*beta0*lumi%l00*pi + L*lumi%l00*A(1)))*B(1)**2 + 2*L**2*lumi%l00*B(1)**3 + 4*beta0*L*lumi%l00*pi*(4*L*A(2) + 3*B(2)) + B(1)*(3*lumi%l20 + 2*L*(-3*lumi%l21 + 3*pi*(beta0*lumi%l10 + 2*beta1*lumi%l00*pi) + L*(6*lumi%l22 + 2*beta0*pi*(-3*lumi%l11 + 4*beta0*lumi%l00*pi)) + L*(-3*lumi%l10 + 6*L*lumi%l11 - 14*beta0*L*lumi%l00*pi)*A(1) + 3*L**3*lumi%l00*A(1)**2) - 6*L**2*lumi%l00*A(2) - 6*L*lumi%l00*B(2)) + 3*(lumi%l31 + L*(-2*lumi%l32 + 4*L*lumi%l33 + (lumi%l10 - 2*L*lumi%l11)*A(2)) + (lumi%l10 - 2*L*lumi%l11)*B(2)) + L*A(1)*(3*lumi%l20 + 2*L*(-3*lumi%l21 + 6*L*lumi%l22 + 4*beta0*lumi%l10*pi - 8*beta0*L*lumi%l11*pi + 8*beta1*lumi%l00*pi**2 + 12*beta0**2*L*lumi%l00*pi**2 - 3*L*lumi%l00*A(2) - 3*lumi%l00*B(2)))) + 2*(9*mu*zeta5*A(1)**2*(lumi%l11 + 2*L*lumi%l00*A(1) + lumi%l00*B(1)) + zeta3*(4*delta*pi*(beta0*lumi%l10 + 2*beta1*lumi%l00*pi - 2*beta0*L*(lumi%l11 - 6*beta0*lumi%l00*pi))*A(1) + 2*L*(-3*gamma*lumi%l10 + L*((6 + 9*gamma)*lumi%l11 - 4*beta0*(delta + 9*gamma)*lumi%l00*pi))*A(1)**2 + 4*(2 + 3*gamma)*L**3*lumi%l00*A(1)**3 - 3*gamma*lumi%l11*A(2) + 8*beta0*delta*lumi%l00*pi*A(2) + (8*beta0*delta*lumi%l00*pi*(beta0*pi - L*A(1)) + 6*(lumi%l22 + 2*L*A(1)*(lumi%l11 + L*lumi%l00*A(1))) - 3*gamma*(2*beta0*pi*(lumi%l11 + 8*L*lumi%l00*A(1)) + A(1)*(lumi%l10 - 4*L*lumi%l11 - 6*L**2*lumi%l00*A(1)) + lumi%l00*A(2)))*B(1) + 3*(lumi%l11 - 2*beta0*gamma*lumi%l00*pi + 2*(1 + gamma)*L*lumi%l00*A(1))*B(1)**2 + lumi%l00*B(1)**3 - 3*A(1)*(-4*L*(lumi%l22 + gamma*(lumi%l22 - 2*beta0*lumi%l11*pi - lumi%l00*A(2))) + gamma*(lumi%l21 + lumi%l00*B(2))))) - 6*L*lumi%l00*B(3))/3.

         elseif (cs%observable == 'phi') then

            resummed_expanded(0,:) = lumi%l00

            resummed_expanded(1,:) = 0.16666666666666666*(0. + 6.*lumi%l10 + 8.317766166719343*lumi%l11 - 12.*L*lumi%l11 - 15.635040568107774*lumi%l00*A(1) + 16.635532333438686*L*lumi%l00*A(1) - 12.*L**2*lumi%l00*A(1) + 8.317766166719343*lumi%l00*B(1) - 12.*L*lumi%l00*B(1))

            resummed_expanded(2,:) = 39.47841760435743*(0.1320134013880541*lumi%l22 + 0.025330295910584444*(lumi%l20 - 2.*L*lumi%l21 + 4.*L**2*lumi%l22) + 1.8875396716123116*beta0*lumi%l00*A(1) - 0.4506169035287284*lumi%l11*A(1) + 0.8601635316621348*lumi%l00*A(1)**2 + 0.1320134013880541*lumi%l11*(2.*L*A(1) + B(1)) - 0.45061690352872846*lumi%l00*A(1)*(2.*L*A(1) + B(1)) + 0.06600670069402703*lumi%l00*(2.*L*A(1) + B(1))**2 + 0.05066059182116889*(lumi%l10 - 2.*L*lumi%l11)*(0. - 1.*L**2*A(1) - 1.*L*B(1)) + 0.03511524638634144*(0. + lumi%l21 - 4.*L*lumi%l22 - 2.*L**2*lumi%l11*A(1) - 2.*L*lumi%l11*B(1)) - 0.06600670069402705*(0. + 25.132741228718345*beta0*L*lumi%l00*A(1) + lumi%l10*A(1) - 2.*L*lumi%l11*A(1) - 2.*L**2*lumi%l00*A(1)**2 + lumi%l00*A(2) + 6.283185307179586*beta0*lumi%l00*B(1) - 2.*L*lumi%l00*A(1)*B(1)) + 0.6931471805599453*(0.3183098861837907*(2.*L*A(1) + B(1))*(0.15915494309189535*(lumi%l10 - 2.*L*lumi%l11) + 0.3183098861837907*lumi%l00*(0. - 1.*L**2*A(1) - 1.*L*B(1))) + 0.05066059182116889*lumi%l00*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2))) + 0.5*lumi%l00*(0.10132118364233778*(0. - 1.*L**2*A(1) - 1.*L*B(1))**2 + 0.03377372788077926*(0. - 25.132741228718345*beta0*L**3*A(1) - 3.*L**2*A(2) - 18.84955592153876*beta0*L**2*B(1) - 3.*L*B(2))))

            resummed_expanded(3,:) = 248.05021344239853*(0.004031441804149937*(lumi%l30 - 2.*L*lumi%l31 + 4.*L**2*lumi%l32 - 8.*L**3*lumi%l33) - 0.5475971117256431*lumi%l22*A(1) + 1.3841580313432646*lumi%l11*A(1)**2 + 0.14343581527472174*lumi%l22*(2.*L*A(1) + B(1)) + 2.2937694177656933*beta0*lumi%l00*A(1)*(2.*L*A(1) + B(1)) - 0.5475971117256432*lumi%l11*A(1)*(2.*L*A(1) + B(1)) + 1.3841580313432646*lumi%l00*A(1)**2*(2.*L*A(1) + B(1)) + 0.07171790763736087*lumi%l11*(2.*L*A(1) + B(1))**2 - 0.2737985558628216*lumi%l00*A(1)*(2.*L*A(1) + B(1))**2 + 0.023905969212453627*lumi%l00*(2.*L*A(1) + B(1))**3 + 0.008062883608299874*(lumi%l20 - 2.*L*lumi%l21 + 4.*L**2*lumi%l22)*(0. - 1.*L**2*A(1) - 1.*L*B(1)) + 0.021010585385283286*(0. + lumi%l32 - 6.*L*lumi%l33 - 2.*L**2*lumi%l22*A(1) - 2.*L*lumi%l22*B(1)) + 0.30041126901915227*(0. + 6.283185307179586*beta1*lumi%l00*A(1) + 37.69911184307752*beta0**2*L*lumi%l00*A(1) + beta0*lumi%l10*A(1) - 2.*beta0*L*lumi%l11*A(1) - 2.*beta0*L**2*lumi%l00*A(1)**2 + 2.*beta0*lumi%l00*A(2) + 6.283185307179586*beta0**2*lumi%l00*B(1) - 2.*beta0*L*lumi%l00*A(1)*B(1)) - 0.07171790763736087*(0. + 25.132741228718345*beta0*L*lumi%l11*A(1) + lumi%l21*A(1) - 4.*L*lumi%l22*A(1) - 2.*L**2*lumi%l11*A(1)**2 + lumi%l11*A(2) + 6.283185307179586*beta0*lumi%l11*B(1) - 2.*L*lumi%l11*A(1)*B(1)) + 0.045633092643803594*(0. + 50.26548245743669*beta0*lumi%l11*A(1) + 150.79644737231007*beta0*L*lumi%l00*A(1)**2 + 3.*lumi%l10*A(1)**2 - 6.*L*lumi%l11*A(1)**2 - 6.*L**2*lumi%l00*A(1)**3 + 6.*lumi%l00*A(1)*A(2) + 37.69911184307752*beta0*lumi%l00*A(1)*B(1) - 6.*L*lumi%l00*A(1)**2*B(1)) + 0.06600670069402703*(2.*L*A(1) + B(1))**2*(0.15915494309189535*(lumi%l10 - 2.*L*lumi%l11) + 0.3183098861837907*lumi%l00*(0. - 1.*L**2*A(1) - 1.*L*B(1))) + 0.021010585385283286*lumi%l00*(2.*L*A(1) + B(1))*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2)) + 0.010505292692641643*(0.3183098861837907*(2.*L*A(1) + B(1))*(6.283185307179586*lumi%l21 - 25.132741228718345*L*lumi%l22 + 12.566370614359172*lumi%l11*(0. - 1.*L**2*A(1) - 1.*L*B(1))) + 2.*lumi%l11*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2))) + 0.125*(-5.6626190148369355*(2.*L*A(1) + B(1))*(0.10132118364233778*lumi%l00*(0. + 25.132741228718345*beta0*L*A(1) + A(2) + 6.283185307179586*beta0*B(1)) + 0.6366197723675814*A(1)*(0.15915494309189535*(lumi%l10 - 2.*L*lumi%l11) + 0.3183098861837907*lumi%l00*(0. - 1.*L**2*A(1) - 1.*L*B(1)))) - 0.573743261098887*lumi%l00*A(1)*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2))) + 0.07957747154594767*(lumi%l10 - 2.*L*lumi%l11)*(0.10132118364233778*(0. - 1.*L**2*A(1) - 1.*L*B(1))**2 + 0.03377372788077926*(0. - 25.132741228718345*beta0*L**3*A(1) - 3.*L**2*A(2) - 18.84955592153876*beta0*L**2*B(1) - 3.*L*B(2))) + 0.005588765040276056*(lumi%l31 - 4.*L*lumi%l32 + 12.*L**2*lumi%l33 + 0.3183098861837907*(6.283185307179586*lumi%l21 - 25.132741228718345*L*lumi%l22)*(0. - 1.*L**2*A(1) - 1.*L*B(1)) + 19.739208802178716*lumi%l11*(0.10132118364233778*(0. - 1.*L**2*A(1) - 1.*L*B(1))**2 + 0.03377372788077926*(0. - 25.132741228718345*beta0*L**3*A(1) - 3.*L**2*A(2) - 18.84955592153876*beta0*L**2*B(1) - 3.*L*B(2)))) + 0.6514600236711572*(-0.10132118364233778*(0. + 25.132741228718345*beta0*L*A(1) + A(2) + 6.283185307179586*beta0*B(1))*(0.15915494309189535*(lumi%l10 - 2.*L*lumi%l11) + 0.3183098861837907*lumi%l00*(0. - 1.*L**2*A(1) - 1.*L*B(1))) - 0.016125767216599748*lumi%l00*(0. + 157.91367041742973*beta1*L*A(1) + 473.7410112522892*beta0**2*L**2*A(1) + 50.26548245743669*beta0*L*A(2) + A(3) + 39.47841760435743*beta1*B(1) + 157.91367041742973*beta0**2*L*B(1) + 12.566370614359172*beta0*B(2)) - 0.6366197723675814*A(1)*(0.025330295910584444*(lumi%l20 - 2.*L*lumi%l21 + 4.*L**2*lumi%l22) + 0.05066059182116889*(lumi%l10 - 2.*L*lumi%l11)*(0. - 1.*L**2*A(1) - 1.*L*B(1)) + 0.5*lumi%l00*(0.10132118364233778*(0. - 1.*L**2*A(1) - 1.*L*B(1))**2 + 0.03377372788077926*(0. - 25.132741228718345*beta0*L**3*A(1) - 3.*L**2*A(2) - 18.84955592153876*beta0*L**2*B(1) - 3.*L*B(2))))) + 0.6931471805599453*(0.05066059182116889*(0.15915494309189535*(lumi%l10 - 2.*L*lumi%l11) + 0.3183098861837907*lumi%l00*(0. - 1.*L**2*A(1) - 1.*L*B(1)))*(0. + 25.132741228718345*beta0*L**2*A(1) + 2.*L*A(2) + 12.566370614359172*beta0*L*B(1) + B(2)) + 0.3183098861837907*(2.*L*A(1) + B(1))*(0.025330295910584444*(lumi%l20 - 2.*L*lumi%l21 + 4.*L**2*lumi%l22) + 0.05066059182116889*(lumi%l10 - 2.*L*lumi%l11)*(0. - 1.*L**2*A(1) - 1.*L*B(1)) + 0.5*lumi%l00*(0.10132118364233778*(0. - 1.*L**2*A(1) - 1.*L*B(1))**2 + 0.03377372788077926*(0. - 25.132741228718345*beta0*L**3*A(1) - 3.*L**2*A(2) - 18.84955592153876*beta0*L**2*B(1) - 3.*L*B(2)))) + 0.008062883608299874*lumi%l00*(0. + 157.91367041742973*beta1*L**2*A(1) + 315.82734083485946*beta0**2*L**3*A(1) + 50.26548245743669*beta0*L**2*A(2) + 2.*L*A(3) + 78.95683520871486*beta1*L*B(1) + 157.91367041742973*beta0**2*L**2*B(1) + 25.132741228718345*beta0*L*B(2) + B(3))) + 0.3333333333333333*lumi%l00*(0.010750511477733165*(0. - 1.*L**2*A(1) - 1.*L*B(1))*(0. - 25.132741228718345*beta0*L**3*A(1) - 3.*L**2*A(2) - 18.84955592153876*beta0*L**2*B(1) - 3.*L*B(2)) + 0.15915494309189535*(0. - 1.*L**2*A(1) - 1.*L*B(1))*(0.10132118364233778*(0. - 1.*L**2*A(1) - 1.*L*B(1))**2 + 0.03377372788077926*(0. - 25.132741228718345*beta0*L**3*A(1) - 3.*L**2*A(2) - 18.84955592153876*beta0*L**2*B(1) - 3.*L*B(2))) + 3.*(-0.6366197723675814*beta0**2*L**4*A(1) - 0.4244131815783876*(0. + beta1*L**3*A(1)) - 0.13509491152311703*beta0*L**3*A(2) - 0.4244131815783876*beta0**2*L**3*B(1) - 0.3183098861837907*(0. + 0.025330295910584444*L**2*A(3) + beta1*L**2*B(1) + 0.3183098861837907*beta0*L**2*B(2)) - 0.008062883608299874*(0. + L*B(3)))))

         end if
      end if

      resummed_expanded(1,:) = resummed_expanded(1,:) * (cs%alphas_muR/twopi)
      resummed_expanded(2,:) = resummed_expanded(2,:) * (cs%alphas_muR/twopi)**2
      resummed_expanded(3,:) = resummed_expanded(3,:) * (cs%alphas_muR/twopi)**3

   end subroutine analytic_expansion



   subroutine expanded_sigma_ptj(v, order, msqB, msqV1, msqV2, x1, x2, resummed_expanded, lumi_asympt)
      implicit none
      real(dp),                  intent(in) :: v(:)
      real(dp), intent(in) :: x1, x2
      integer,                   intent(in) :: order !(0=LL,1=NLL,2=NNLL)
      real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int),msqV1(-nf_int:nf_int,-nf_int:nf_int),msqV2(-nf_int:nf_int,-nf_int:nf_int)
      ! first index of res denotes the order in alphas
      real(dp), intent(out) :: resummed_expanded(0:3,size(v))
      real(dp), optional, intent(out) :: lumi_asympt(0:3)
      type(PDFExpansion_at_x)  :: pdf1eatx, pdf2eatx
      type(LumiExpansion_at_x) :: lumi

      if (order < order_NLL) return ! use standard resummation for LL

      ! evaluate the pdf expansions at the Bjorken variables given by the Born event
      call set_pdf_expansion_at_x(pdf1eatx, pdf2eatx, pdf1e, pdf2e, x1, x2)

      !--- the following call returns luminosities factors lij, where lij is
      !--- the coefficient of (as2pi(muR))^i * t^j with  t = 2*log(muF1/muF0)=-2*L
      call set_lumi_expansion_Born(pdf1eatx, pdf2eatx, msqB, msqV1, msqV2, lumi, order)

      ! Note that since we are using the analytic expansion there is only one lumi
      !This differs from standard radish implementation
      call ptj_expansion(v,order,lumi,resummed_expanded)

      if (present(lumi_asympt)) then
         lumi_asympt = zero ! asymptotic limit for the multiplicative matching
         ! For multiplicative matching scheme, return asymptotic normalization
         if (order == order_NLL)  lumi_asympt(0)  = lumi_NLL_Born(cs%muF, zero, x1, x2, msqB, msqV1)
         if (order == order_NNLL) lumi_asympt(0)  = lumi_NNLL_Born(cs%muF, zero, x1, x2, msqB, msqV1, msqV2)
         if (order == order_NNNLL) lumi_asympt(0) = lumi_NNNLL_Born(cs%muF, zero, x1, x2, msqB, msqV1, msqV2)
      end if

      ! For multiplicative matching scheme, return the expanded counterterm up to O(as^3). Asymptotic limit goes like:
      if (present(lumi_asympt)) then
         lumi_asympt(1) = lumi%l10 * (cs%alphas_muR/twopi)
         lumi_asympt(2) = lumi%l20 * (cs%alphas_muR/twopi)**2
         lumi_asympt(3) = lumi%l30 * (cs%alphas_muR/twopi)**3
      end if

      ! Hoppet returns x*f(x) - restore correct normalisation
      resummed_expanded(:,:) = resummed_expanded(:,:)/x1/x2
      lumi_asympt(1:) = lumi_asympt(1:)/x1/x2

   end subroutine expanded_sigma_ptj

   subroutine expanded_sigma_ptj_inclusive(v, order, resummed_expanded, lumi_asympt)
      implicit none
      real(dp),                  intent(inout) :: v(:)
      integer,                   intent(inout) :: order !(0=LL,1=NLL,2=NNLL)
      type(LumiExpansion) :: lumie!, luminll, luminnll
      type(LumiExpansion_at_x) :: lumi
      ! first index of res denotes the order in alphas
      real(dp), intent(out) :: resummed_expanded(0:3,size(v))
      real(dp), optional, intent(out) :: lumi_asympt(0:3)

      if (order < order_NLL) return ! use standard resummation for LL

      call AllocLumiExpansion(lumie)

      !--- the following call returns luminosities factors lij, where lij is
      !--- the coefficient of (as2pi(muR))^i * t^j with  t = 2*log(muF1/muF0)=-2*L
      call inclusive_set_lumi_expansion_ptB(pdf1e,pdf2e,lumie, order)
      !--- evaluate luminosity at x=M^2/rts^2
      call inclusive_set_lumi_expansion_at_x(lumi,lumie)

      call ptj_expansion(v,order,lumi,resummed_expanded)

      if (present(lumi_asympt)) then
         lumi_asympt = zero ! asymptotic limit for the multiplicative matching
         ! For multiplicative matching scheme, return asymptotic normalization
         if (order == order_NLL)  lumi_asympt(0)  = inclusive_lumi_NLL_loc(cs%muF, zero)
         if (order == order_NNLL) lumi_asympt(0)  = inclusive_lumi_NNLL_loc(cs%muF, zero)
         if (order == order_NNNLL) lumi_asympt(0) = inclusive_lumi_NNNLL_loc(cs%muF, zero)
      end if

      ! For multiplicative matching scheme, return the expanded counterterm up to O(as^3). Asymptotic limit goes like:
      if (present(lumi_asympt)) then
         lumi_asympt(1) = lumi%l10 * (cs%alphas_muR/twopi)
         lumi_asympt(2) = lumi%l20 * (cs%alphas_muR/twopi)**2
         lumi_asympt(3) = lumi%l30 * (cs%alphas_muR/twopi)**3
      end if

      call Delete(lumie)

   end subroutine expanded_sigma_ptj_inclusive

   !Analytical expansion adapted from ptveto
   subroutine ptj_expansion(v, order, lumi, resummed_expanded)
      implicit none
      real(dp),                  intent(in) :: v(:)
      integer,                   intent(in) :: order !(0=LL,1=NLL,2=NNLL)
      type(LumiExpansion_at_x), intent(in) :: lumi!, luminll, luminnll
      ! first index of res denotes the order in alphas
      real(dp), intent(out) :: resummed_expanded(0:3,size(v))
      !-------------------------------------
      real(dp) :: L(size(v)), lambda(size(v))
      real(dp) :: h11, h12, h21, h22, h23, h24
      real(dp) :: h36, h35, h34, h33, h32, h31
      integer  :: i
      real(dp) :: t_array(size(v))

      L = Ltilde(v/cs%Q, cs%p)
      t_array = v/cs%Q
      lambda  = get_lambda(L)


      resummed_expanded = zero

      ! Born cross section
      resummed_expanded(0,:) = lumi%l00

      !-- LL terms
      h12 = -two**2*CC
      h11 = zero
      h21 = zero
      h22 = zero
      h23 = -32._dp*CC*Pi*beta0/(3._dp)
      h24 = two**3*CC**2
      h36 = -32._dp/3._dp*CC**3
      h35 = 128._dp/3._dp*CC**2*beta0*Pi
      h34 = -32._dp*CC*beta0**2*Pi**2
      h33 = zero
      h32 = zero
      ! - Added by FD -
      h31 = zero

      !-- NLL terms
      if (order >= order_NLL) then

         h11 = h11 - 8._dp*BB*CC+4._dp*CC*cs%ln_Q2_M2
         h22 = h22 + 32*BB**2*CC**2-16*BB*CC*beta0*Pi-32*BB*CC**2*cs%ln_Q2_M2 &
            & + 16*CC*beta0*Pi*cs%ln_Q2_M2+8*CC**2*cs%ln_Q2_M2**2-8*CC*beta0*Pi*cs%ln_muR2_M2 &
            & - 4*CC*cmw_K
         h23 = h23 + 32*BB*CC**2-16*CC**2*cs%ln_Q2_M2
         h35 = h35 - 64._dp*BB*CC**3+32._dp*CC**3*cs%ln_Q2_M2
         h34 = h34 -128*BB**2*CC**3 + (448*BB*beta0*CC**2*Pi)/3._dp + 128*BB*CC**3*cs%ln_Q2_M2 &
            & - (224*beta0*CC**2*Pi*cs%ln_Q2_M2)/3._dp - 32*CC**3*cs%ln_Q2_M2**2 &
            & + 32*beta0*CC**2*Pi*(cs%ln_muR2_M2-cs%ln_Q2_M2) + 16*CC**2*cmw_K

         h33 = h33 + (-256*BB**3*CC**3)/3._dp + 128*BB**2*beta0*CC**2*Pi - (128*BB*beta0**2*CC*Pi**2)/3._dp &
            & - (64*beta1*CC*Pi**2)/3._dp + 128*BB**2*CC**3*cs%ln_Q2_M2 &
            & - 128*BB*beta0*CC**2*Pi*cs%ln_Q2_M2 + (64*beta0**2*CC*Pi**2*cs%ln_Q2_M2)/3._dp &
            & - 64*BB*CC**3*cs%ln_Q2_M2**2 + 32*beta0*CC**2*Pi*cs%ln_Q2_M2**2 &
            & + (32*CC**3*cs%ln_Q2_M2**3)/3._dp +  64*BB*beta0*CC**2*Pi*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & - (128*beta0**2*CC*Pi**2*(cs%ln_muR2_M2-cs%ln_Q2_M2))/3._dp &
            & - 32*beta0*CC**2*Pi*cs%ln_Q2_M2*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & + 32*BB*CC**2*cmw_K - (64*beta0*CC*Pi*cmw_K)/3._dp &
            & - 16*CC**2*cs%ln_Q2_M2*cmw_K

      end if

      ! NNLL terms
      if (order > order_NLL) then
         h21 = h21 - two*B(2)-16*BB*CC*beta0*Pi*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & +8*CC*beta0*Pi*cs%ln_Q2_M2*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
         & +4._dp*CC*cmw_K*cs%ln_Q2_M2

         h21 = h21 + 16*CC*non_incl(cs%jet_radius,'all')

         h33 = h33 + 8*B(2)*CC + 64*BB*beta0*CC**2*Pi*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & - 32*beta0*CC**2*Pi*cs%ln_Q2_M2*(cs%ln_muR2_M2-cs%ln_Q2_M2)  &
            & - 16*CC**2*cs%ln_Q2_M2*cmw_K - 64*CC**2*non_incl(cs%jet_radius,'all')

         h32 = h32 - 2*A(3) + 16*B(2)*BB*CC - 8*B(2)*beta0*Pi - 32*BB*beta1*CC*Pi**2 &
            & - 8*B(2)*CC*cs%ln_Q2_M2 + 16*beta1*CC*Pi**2*cs%ln_Q2_M2 &
            & + 128*BB**2*beta0*CC**2*Pi*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & - 64*BB*beta0**2*CC*Pi**2*(cs%ln_muR2_M2-cs%ln_Q2_M2)  &
            & - 16*beta1*CC*Pi**2*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & - 128*BB*beta0*CC**2*Pi*cs%ln_Q2_M2*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & + 32*beta0**2*CC*Pi**2*cs%ln_Q2_M2*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & + 32*beta0*CC**2*Pi*cs%ln_Q2_M2**2*(cs%ln_muR2_M2-cs%ln_Q2_M2) &
            & - 16*beta0**2*CC*Pi**2*(cs%ln_muR2_M2-cs%ln_Q2_M2)**2 &
            & - 32*BB*CC**2*cs%ln_Q2_M2*cmw_K + 16*beta0*CC*Pi*cs%ln_Q2_M2*cmw_K &
            & + 16*CC**2*cs%ln_Q2_M2**2*cmw_K - 16*beta0*CC*Pi*(cs%ln_muR2_M2-cs%ln_Q2_M2)*cmw_K &
            & - 128*BB*CC**2*non_incl(cs%jet_radius,'all') &
            & + 128*beta0*CC*Pi*non_incl(cs%jet_radius,'all') &
            & + 64*CC**2*cs%ln_Q2_M2*non_incl(cs%jet_radius,'all')
      end if

      !---------------------------------------------------------------------------------

      ! Build the expansion
      !-- LL terms
      resummed_expanded(1,:) = lumi%l00*h12*L**2
      resummed_expanded(2,:) = lumi%l00*(h24*L**4 + h23*L**3)
      resummed_expanded(3,:) = lumi%l00*(h36*L**6 + h35*L**5 + h34*L**4)
      !-- NLL and NNLL terms
      if (order >= order_NLL) then
         resummed_expanded(1,:) = resummed_expanded(1,:) + (lumi%l00*h11 - two*lumi%l11)*L + lumi%l10
         resummed_expanded(2,:) = resummed_expanded(2,:) + (-two*lumi%l11*h12)*L**3 &
            & + (lumi%l00*h22 + lumi%l10*h12 - two*lumi%l11*h11 + four*lumi%l22)*L**2 &
            & + (lumi%l00*h21 + lumi%l10*h11 - two*lumi%l21)*L  + lumi%l20

         resummed_expanded(3,:) = resummed_expanded(3,:) + (-two*lumi%l11*h24)*L**5 &
            & + (lumi%l10*h24 - two*lumi%l11*h23 + four*lumi%l22*h12)*L**4 &
            & + (lumi%l00*h33 + lumi%l10*h23 - two*lumi%l11*h22 - two*lumi%l21*h12 &
            & + four*lumi%l22*h11 - two*four*lumi%l33)*L**3 &
            & + (lumi%l00*h32 + lumi%l10*h22 - two*lumi%l11*h21 + lumi%l20*h12 &
            & - two*lumi%l21*h11 + four*lumi%l32)*L**2 &
            & + (lumi%l00*h31 + lumi%l10*h21 + lumi%l20*h11 - two*lumi%l31)*L + lumi%l30
      end if

      resummed_expanded(1,:) = resummed_expanded(1,:) * (cs%alphas_muR/twopi)
      resummed_expanded(2,:) = resummed_expanded(2,:) * (cs%alphas_muR/twopi)**2
      resummed_expanded(3,:) = resummed_expanded(3,:) * (cs%alphas_muR/twopi)**3
   end subroutine ptj_expansion

   subroutine terminate_expansion
      implicit none
      call Delete(pdf1e)
      call Delete(pdf2e)
   end subroutine terminate_expansion

end module expansion
