! This module contains the ingredients to compute the top/bottom-quark
! mass corrections in the hard virtual coefficient to NLO
module mass_corr
  use types; use consts_dp
  use ew_parameters
  use hoppet_v1
  implicit none
  private

  public :: c1h, norm_mass

contains
  !======================================================================
  ! running quark mass as defined in Spira, Djouadi, Graudenz & Zerwas
  ! hep-ph/9504378
  ! Different from usual MSbar mass
  function mq(mpole,mu,scheme) result(res)
    real(dp), intent(in) :: mpole, mu
    real(dp)             :: res
    character(len=4), intent(in)     :: scheme
    real(dp) :: x, C_mu, C_mu0

    !scheme='M_OS' ! for the on-shell running mass
    !scheme='M_MS' ! for the MSbar mass at mu
    !---------------------------------------------
    if (scheme=='M_OS') then
       if (mu > half) then ! put a cutoff at 500 MeV (for the strange quark)
          res = mpole*(RunningCoupling(mu)/RunningCoupling(mpole))**(one/pi/beta0)
       else
          res = mpole
       end if
    elseif (scheme=='M_MS') then
       if (mu > half.and.mpole > half) then
          ! result is from eq. (29) of Chetyrkin, Steinhauser 9911434, but it dates much back
          ! Here the running coupling is evaluated at nf=5, perhaps for light quarks this should be changed.
          ! convert from on-shell to MSbar at mpole.
          ! In the following formula mpole actually denotes the MSbar mass mQ(mQ) (with mQ being the initial condition in the MSbar scheme)
          ! Run at mu in the MSbar scheme (eq. (9) of (0004189))
          res = mpole * C(RunningCoupling(mu)/Pi)/C(RunningCoupling(mpole)/Pi)
       else
          res = mpole
       end if
    end if
  end function mq

  ! Equation (10) of the RunDec paper (0004189). Running mass at three loops in the MSbar scheme
  function C(x) result(res)
    real(dp), intent(in) :: x
    real(dp) :: res
    real(dp) :: gamma0, gamma1, gamma2, gamma3
    real(dp), parameter :: zeta4=Pi**4/90._dp, zeta5=1.036927755143370_dp
    ! MS bar anomalous dimensions in units of (as/Pi)^i
    gamma0=one
    gamma1=one/16._dp*(202._dp/three-20._dp/9._dp*nf)
    gamma2=one/64._dp*(1249._dp + (-2216._dp/27._dp - 160._dp/3._dp*zeta3)*nf - 140._dp/81._dp*nf**2)
    gamma3=one/256._dp*(4603055._dp/162._dp + 135680._dp/27._dp*zeta3-8800._dp*zeta5 + (-91723._dp/27._dp - 34192._dp/9._dp*zeta3 &
         & + 880._dp*zeta4 + 18400._dp/9._dp*zeta5)*nf + (5242._dp/243._dp+800._dp/9._dp*zeta3 - 160._dp/three*zeta4)*nf**2 &
         & + (-332._dp/243._dp+64._dp/27._dp*zeta3)*nf**3)

    res=x**(gamma0/(Pi*beta0))*(one + (gamma1/(Pi*beta0) - Pi*beta1/beta0*gamma0/(Pi*beta0))*x &
               & + half*x**2*((gamma1/(Pi*beta0) - Pi*beta1/beta0*gamma0/(Pi*beta0))**2 + gamma2/(Pi*beta0) &
               & - Pi*beta1/beta0*gamma1/(Pi*beta0) + (Pi*beta1/beta0)**2*gamma0/(Pi*beta0) - Pi**2*beta2/beta0*gamma0/(Pi*beta0)))
  end function C

  !======================================================================
  ! correction factor to account for mass effects in the Born
  ! cross section for Higgs production
  ! expressions from Harlander & Kant hep-ph/0509189
  ! norm_mass is the ratio of the "finite-mass" cross section
  ! to the "heavy-quark" result
  ! Agreement with hnnlo-v2.0 if muq=mpole (q=t,b)
  ! mut is the scale at which the top-quark mass is defined
  ! mub is the scale at which the bottom-quark mass is defined
  function norm_mass(M, mut, mub, muc, mus, case, scheme) result(res)
    character(len=*), intent(in) :: case
    real(dp)        , intent(in) :: M, mut, mub, muc, mus
    character(len=4), intent(in)     :: scheme
    !---------------------------------------------
    real(dp) :: res, taus, tauc, taub, taut

    ! Eq. (2.2)
    taus = M**2/four/mq(ms,mus,scheme)**2
    tauc = M**2/four/mq(mc,muc,scheme)**2
    taub = M**2/four/mq(mb,mub,scheme)**2
    taut = M**2/four/mq(mt,mut,scheme)**2

    select case(trim(case))
    case('t')
       res = abs(f0(taut))**2 ! the ratio between masses rescales the Yukawa coupling
    case('b')
       res = abs(yb*f0(taub))**2
    case('t+b')
       res = abs(f0(taut) + yb*f0(taub))**2
    case('t+b+c')
       res = abs(f0(taut) + yb*f0(taub) + yc*f0(tauc))**2
    case('t+b+c+s')
       res = abs(f0(taut) + yb*f0(taub) + yc*f0(tauc) + ys*f0(taus))**2
    case default
       call wae_error('heavy-quark mass corrections:  unrecognized case: '//case)
    end select
  end function norm_mass

  !======================================================================
  ! two-loop virtual corrections with heavy quarks running in the loop
  ! expressions from Harlander & Kant hep-ph/0509189
  ! Eqs. (3.5,3.6)
  ! mut is the scale at which the top-quark mass is defined
  ! mub is the scale at which the bottom-quark mass is defined
  function c1h(M,mut,mub,muc,mus,case,scheme) result(res)
    character(len=*), intent(in) :: case
    real(dp)        , intent(in) :: M, mut, mub, muc, mus
    character(len=4), intent(in)     :: scheme
    !---------------------------------------------
    real(dp) :: res, taus, tauc, taub, taut

    taus = M**2/four/mq(ms,mus,scheme)**2
    tauc = M**2/four/mq(mc,muc,scheme)**2
    taub = M**2/four/mq(mb,mub,scheme)**2
    taut = M**2/four/mq(mt,mut,scheme)**2

    res = zero

    select case (scheme)
    case ('M_OS')   ! implement on-shell scheme as in 0611266, Eq. (15)
       select case(trim(case))
       case('t')
          res = real((f12a(taut) + two*f0c2(taut)*(-four/three))/f0tau(taut))
       case('b')
          res = real((f12a(taub) + two*f0c2(taub)*(-four/three))/f0tau(taub))
       case('t+b')
          res = real((f12a(taut) + two*f0c2(taut)*(-four/three) &
               &    + yb*f12a(taub) + yb*two*f0c2(taub)*(-four/three))/(f0tau(taut)+yb*f0tau(taub)))
       case('t+b+c')
          res = real((f12a(taut) + two*f0c2(taut)*(-four/three) &
               &    + yb*f12a(taub) + yb*two*f0c2(taub)*(-four/three) &
               &    + yc*f12a(tauc) + yc*two*f0c2(tauc)*(-four/three)) &
               &    /(f0tau(taut) + yb*f0tau(taub) + yc*f0tau(tauc)))
       case('t+b+c+s')
          res = real((f12a(taut) + two*f0c2(taut)*(-four/three) &
               &    + yb*f12a(taub) + yb*two*f0c2(taub)*(-four/three) &
               &    + yc*f12a(tauc) + yc*two*f0c2(tauc)*(-four/three) &
               &    + ys*f12a(taus) + ys*two*f0c2(taus)*(-four/three)) &
               &    /(f0tau(taut) + yb*f0tau(taub) + yc*f0tau(tauc) + ys*f0tau(taus)))
       case default
          call wae_error('heavy-quark mass corrections:  unrecognized case: '//case)
       end select

    case ('M_MS')  ! implement on-shell scheme as in 0611266, Eq. (11)
       select case(trim(case))
       case('t')
          res = real((f12a(taut) + two*f0c2(taut)*log(mut**2/mq(mt,mut,scheme)**2))/f0tau(taut))
       case('b')
          res = real((f12a(taub) + two*f0c2(taub)*log(mub**2/mq(mb,mub,scheme)**2))/f0tau(taub))
       case('t+b')
          res = real((f12a(taut) + two*f0c2(taut)*log(mut**2/mq(mt,mut,scheme)**2) &
               &    + yb*f12a(taub) + yb*two*f0c2(taub)*log(mub**2/mq(mb,mub,scheme)**2))/(f0tau(taut)+yb*f0tau(taub)))
       case('t+b+c')
          res = real((f12a(taut) + two*f0c2(taut)*log(mut**2/mq(mt,mut,scheme)**2) &
               &    + yb*f12a(taub) + yb*two*f0c2(taub)*log(mub**2/mq(mb,mub,scheme)**2) &
               &    + yc*f12a(tauc) + yc*two*f0c2(tauc)*log(muc**2/mq(mc,muc,scheme)**2)) &
               &    /(f0tau(taut) + yb*f0tau(taub) + yc*f0tau(tauc)))
       case('t+b+c+s')
          res = real((f12a(taut) + two*f0c2(taut)*log(mut**2/mq(mt,mut,scheme)**2) &
               &    + yb*f12a(taub) + yb*two*f0c2(taub)*log(mub**2/mq(mb,mub,scheme)**2) &
               &    + yc*f12a(tauc) + yc*two*f0c2(tauc)*log(muc**2/mq(mc,muc,scheme)**2) &
               &    + ys*f12a(taus) + ys*two*f0c2(taus)*log(mus**2/mq(ms,mus,scheme)**2)) &
               &    /(f0tau(taut) + yb*f0tau(taub) + yc*f0tau(tauc) + ys*f0tau(taus)))
       case default
          call wae_error('heavy-quark mass corrections:  unrecognized case: '//case)
       end select
    case default
       call wae_error("Two-loop virtuals : unknown mass scheme")
    end select

    res = res + pisq

  end function c1h

  !======================================================================
  ! Eq. (2.4)
  function f(tau) result(res)
    real(dp), intent(in) :: tau
    complex(dp) :: res

    if(tau < one) then
       res = asin(sqrt(tau))**2
    else
       res = -one/four*(log((one+sqrt(one-one/tau))/(one-sqrt(one-one/tau)))-dcmplx(zero,one)*pi)**2
    endif

  end function f
  !======================================================================
  ! Derivative of Eq. (2.4)
  function f_p(tau) result(res)
    real(dp), intent(in) :: tau
    complex(dp) :: res

    if (tau < one) then
       res = asin(sqrt(tau))/sqrt(tau*(one-tau))
    else
       res = -half*(log((one+sqrt(one-one/tau))/(one-sqrt(one-one/tau)))-dcmplx(zero,one)*pi)&
            & *(one-sqrt(one-one/tau))/(one+sqrt(one-one/tau))&
            & *one/((sqrt(tau-one)-sqrt(tau))**2*sqrt(tau*(tau-one)))
    endif

  end function f_p
  !======================================================================
  ! Eq. (2.10)
  function theta(tau) result(res)
    real(dp), intent(in) :: tau
    complex(dp) :: res

!    res = (sqrt(one-one/tau)-one)/(sqrt(one-one/tau)+one)
    if (tau < one) then
       res =dcmplx(one-two*tau,two*sqrt(tau*(one-tau)))
    else
       res = dcmplx((sqrt(tau-one)-sqrt(tau))/(sqrt(tau-one)+sqrt(tau)), zero)
    end if
  end function theta


  function f0b1(t) result(res)
    complex(dp), intent(in) :: t
    complex(dp)          :: res
    complex(dp) :: HPL1, HPL2, HPL3, HPL4


    res  =  (-94d0*t)/(1 - t)**2 + (2d0*pi**4*t*(1 + t - 3d0*t**2 -&
         &     3d0*t**3))/&
         &     (5d0*(1 - t)**5) - (24d0*t*(1 + t)*HPL1(0,t))/(1 - t)**3 + &
         &     (3d0*t*(3d0 + 22d0*t + 3d0*t**2)*HPL1(0,t)**2)/(2d0*(1 - t)**4)+ &
         &     (pi**2*t*(1 + t - 17d0*t**2 - 17d0*t**3)*HPL1(0,t)**2)/&
         &     (6d0*(1 - t)**5) + &
         &     (t*(6d0 + 59d0*t + 58d0*t**2 + 33d0*t**3)*HPL1(0,t)**3)/&
         &     (3d0*(1 - t)**5)+ &
         &     (t*(5d0 + 5d0*t - 13d0*t**2 - 13d0*t**3)*HPL1(0,t)**4)&
         &     /(24d0*(1 - t)**5) - &
         &     (t*(47d0 + 66d0*t + 47d0*t**2)*HPL1(0,t)**2*HPL1(1,t))/(1 - t)**4 + &
         &     (16d0*t*(1 + t + t**2 + t**3)*HPL1(0,t)**2*HPL2(0,-1,t))/(1 - t)**5 &
         &     + (2*t*(51d0 + 74d0*t + 51d0*t**2)*HPL1(0,t)*HPL2(0,1,t))/&
         &     (1 - t)**4 - &
         &     (2d0*t*(5d0 + 5d0*t + 23d0*t**2 + 23d0*t**3)*&
         &     HPL1(0,t)**2*HPL2(0,1,t))/(1 - t)**5 - &
         &     (2d0*t*(55d0 + 82d0*t + 55d0*t**2)*HPL3(0,0,1,t))/(1 - t)**4 + &
         &     (4d0*t*(1 + t)*(23d0 + 41d0*t**2)*HPL1(0,t)*&
         &     (-HPL3(0,0,-1,t) + HPL3(0,0,1,t)))/&
         &     (1 - t)**5 + (36d0*t*(5d0 + 5d0*t + 11d0*t**2 + 11d0*t**3)*&
         &     HPL4(0,0,0,-1,t))/&
         &     (1 - t)**5 - (36d0*t*(5d0 + 5d0*t + 7d0*t**2 + 7d0*t**3)*&
         &     HPL4(0,0,0,1,t))/&
         &     (1 - t)**5 + (2d0*t*(31d0 + 34d0*t + 31d0*t**2)*zeta3)/&
         &     (1 - t)**4 + &
         &     (2d0*t*(11d0 + 11d0*t - 43d0*t**2 - 43d0*t**3)*HPL1(0,t)*zeta3)/&
         &     (1 - t)**5 + &
         &     (t*(1 + t)**2*((-4d0*pi**2*HPL1(0,t))/3d0 + &
         &     6d0*pi**2*HPL1(0,t)*HPL1(1,t) - &
         &     6d0*HPL1(0,t)**3*HPL1(1,t) - 32d0*HPL1(0,t)*HPL2(0,-1,t) -&
         &     6d0*pi**2*HPL2(0,1,t) + &
         &     64d0*HPL3(0,0,-1,t) + 72d0*HPL4(1,0,-1,0,t) + 108d0*HPL1(1,t)*zeta3))/&
         &     (1 - t)**4

  end function f0b1

  !======================================================================
  ! Eq. (2.7)
  function f0c2(t) result(res)
    real(dp), intent(in) :: t
    complex(dp)          :: res
    !------------------------------------------

    res = three/t**2*(t+(t-two)*f(t)&
         & - (t-one)*t*f_p(t))
  end function f0c2

  !======================================================================
  ! implement function F12a from Eq. (12) of 0611266
  ! see ancillary file Vicini.nb

  function f12a(tau) result(res)
    real(dp), intent(in) :: tau
    complex(dp) :: res

    !-----
    complex(dp) :: x, HPL1, HPL2, HPL3, HPL4

    res = zero
    x = theta(tau)

    res =  (-2*tau*((4*((36*x)/(-1 + x)**2 - (4*x*(1 - 14*x + x**2)*zeta3)/(-1 + x)**4 - (4*x*(1 + x)*HPL1(0,x))/(-1 + x)**3 - &
         &           (8*x*(1 + 9*x + x**2)*HPL2(0,0,x))/(-1 + x)**4 + (2*x*(3 + 25*x - 7*x**2 + 3*x**3)*HPL3(0,0,0,x))/(-1 + x)**5 + &
         &           (4*x*(1 + 2*x + x**2)*((Pi**2*HPL1(0,x))/6. + 4*HPL3(0,-1,0,x) - HPL3(0,1,0,x)))/(-1 + x)**4 + (4*x*(5 - 6*x + 5*x**2)*HPL3(1,0,0,x))/(-1 + x)**4 -  &
         &           (8*x*(1 + x + x**2 + x**3)*(Pi**4/40. + 2*zeta3*HPL1(0,x) + (Pi**2*HPL2(0,0,x))/6. - 2*HPL4(0,-1,0,0,x) + 4*HPL4(0,0,-1,0,x) + HPL4(0,0,0,0,x)/4. - &
         &                HPL4(0,0,1,0,x) + (7*HPL4(0,1,0,0,x))/2.))/(-1 + x)**5))/3. + &
         &      (12*x*(3 + zeta3 + (x*(1 + 8*x + 3*x**2)*HPL3(0,0,0,x))/(-1 + x)**3 - HPL3(1,0,0,x) - &
         &           (2*(1 + x)**2*(Pi**4/45. + 2*zeta3 + (3*zeta3*HPL1(0,x))/2. + 3*zeta3*HPL1(1,x) + ((1 + Pi**2/3.)*HPL2(0,0,x))/4. + (Pi**2*HPL2(1,0,x))/6. - &
         &                2*HPL3(1,0,0,x) + HPL4(0,0,-1,0,x) + HPL4(0,0,0,0,x)/4. + 2*HPL4(1,0,-1,0,x) - HPL4(1,0,0,0,x)))/(-1 + x)**2))/(-1 + x)**2))/ &
         &  (2 - (1 - 1/tau)*HPL2(0,0,x))

    ! multiply by the Born squared to match my conventions for the two-loop virtuals
    res = res * f0tau(tau)
  end function f12a

  ! Eq. (2.3)
  function f0tau(t) result(res)
    real(dp), intent(in) :: t
    complex(dp)          :: res
    !------------------------------------------

    res = half*three/t**2*(t+(t-one)*f(t))
  end function f0tau


  !======================================================================
  ! Eq. (2.3)
  function f0(t) result(res)
    real(dp), intent(in) :: t
    complex(dp)          :: res
    !------------------------------------------

    res = half*three/t**2*(t+(t-one)*f(t))
   end function f0

end module mass_corr
